#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from datetime import date

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Service, ServiceDate


class ServiceDateTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()
        self.service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2012, 12, 31))

    def test_string(self):
        service_date = ServiceDate.objects.create(
            date=date(2012, 4, 14), service=self.service, exception_type=2)
        self.assertEqual(
            str(service_date), '%d-S1 2012-04-14 Removed' % self.feed.id)
        service_date.exception_type = 1
        self.assertEqual(
            str(service_date), '%d-S1 2012-04-14 Added' % self.feed.id)

    def test_import_calendar_dates_txt(self):
        calendar_dates_txt = StringIO("""\
service_id,date,exception_type
S1,20120414,2
""")
        ServiceDate.import_txt(calendar_dates_txt, self.feed)
        service_date = ServiceDate.objects.get()
        self.assertEqual(service_date.date, date(2012, 4, 14))
        self.assertEqual(service_date.service, self.service)
        self.assertEqual(service_date.exception_type, 2)

    def test_import_calendar_dates_duplicates(self):
        calendar_dates_txt = StringIO("""\
service_id,date,exception_type
S1,20120414,2
S1,20120414,1
""")
        ServiceDate.import_txt(calendar_dates_txt, self.feed)
        service_date = ServiceDate.objects.get()  # Just one
        self.assertEqual(service_date.date, date(2012, 4, 14))
        self.assertEqual(service_date.service, self.service)
        self.assertEqual(service_date.exception_type, 2)

    def test_export_calendar_dates_txt_none(self):
        cdates_txt = ServiceDate.export_txt(self.feed)
        self.assertFalse(cdates_txt)

    def test_export_calendar_dates_txt(self):
        ServiceDate.objects.create(
            date=date(2012, 8, 31), service=self.service, exception_type=2)
        ServiceDate.objects.create(
            date=date(2012, 9, 1), service=self.service, exception_type=1)
        cdates_txt = ServiceDate.export_txt(self.feed)
        self.assertEqual(cdates_txt, """\
service_id,date,exception_type
S1,20120831,2
S1,20120901,1
""")



from django.conf.urls import include, url
from django.views.generic import TemplateView

from django.contrib import admin
admin.autodiscover()

urlpatterns = [
    url(r'^admin/', include(admin.site.urls)),
    url(r'^$', TemplateView.as_view(template_name="explore/home.html"),
        name='home'),
    url(r'', include('exploreapp.urls'))
]



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Agency, Route, Trip


class RouteTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()

    def test_string(self):
        route = Route.objects.create(feed=self.feed, route_id='RTEST', rtype=3)
        self.assertEqual(str(route), '%d-RTEST' % self.feed.id)

    def test_import_routes_txt_minimal(self):
        routes_txt = StringIO("""\
route_id,route_short_name,route_long_name,route_type
AB,10,Airport - Bullfrog,3

""")
        Route.import_txt(routes_txt, self.feed)
        route = Route.objects.get()
        self.assertEqual(route.feed, self.feed)
        self.assertEqual(route.route_id, 'AB')
        self.assertEqual(route.agency, None)
        self.assertEqual(route.short_name, '10')
        self.assertEqual(route.long_name, 'Airport - Bullfrog')
        self.assertEqual(route.desc, '')
        self.assertEqual(route.rtype, 3)
        self.assertEqual(route.url, '')
        self.assertEqual(route.color, '')
        self.assertEqual(route.text_color, '')

    def test_import_routes_txt_duplicate(self):
        routes_txt = StringIO("""\
route_id,route_short_name,route_long_name,route_type
AB,10,Airport - Bullfrog,3
AB,11,Airport Party Bus,3
""")
        Route.import_txt(routes_txt, self.feed)
        route = Route.objects.get()  # Just one
        self.assertEqual(route.feed, self.feed)
        self.assertEqual(route.route_id, 'AB')
        self.assertEqual(route.short_name, '10')

    def test_import_routes_txt_maximal(self):
        routes_txt = StringIO("""\
route_id,agency_id,route_short_name,route_long_name,route_desc,route_type,\
route_url,route_color,route_text_color
AB,DTA,10,Airport - Bullfrog,"Our Airport Route", 3,http://example.com,\
00FFFF,000000
""")
        agency = Agency.objects.create(feed=self.feed, agency_id='DTA')
        Route.import_txt(routes_txt, self.feed)
        route = Route.objects.get()
        self.assertEqual(route.feed, self.feed)
        self.assertEqual(route.route_id, 'AB')
        self.assertEqual(route.agency, agency)
        self.assertEqual(route.short_name, '10')
        self.assertEqual(route.long_name, 'Airport - Bullfrog')
        self.assertEqual(route.desc, 'Our Airport Route')
        self.assertEqual(route.rtype, 3)
        self.assertEqual(route.url, 'http://example.com')
        self.assertEqual(route.color, '00FFFF')
        self.assertEqual(route.text_color, '000000')

    def test_export_routes_txt_none(self):
        routes_txt = Route.export_txt(self.feed)
        self.assertFalse(routes_txt)

    def test_export_routes_txt_minimal(self):
        Route.objects.create(
            feed=self.feed, route_id='AB', short_name='10',
            long_name='Airport - Bullfrog', rtype=3)
        routes_txt = Route.export_txt(self.feed)
        self.assertEqual(routes_txt, """\
route_id,route_short_name,route_long_name,route_type
AB,10,Airport - Bullfrog,3
""")

    def test_export_routes_txt_maximal(self):
        agency = Agency.objects.create(feed=self.feed, agency_id='DTA')
        Route.objects.create(
            feed=self.feed, agency=agency, route_id='AB', short_name='10',
            long_name='Airport - Bullfrog', desc='Our Airport Route',
            rtype=3, url='http://example.com', color='00FFFF',
            text_color='000000')
        routes_txt = Route.export_txt(self.feed)
        self.assertEqual(routes_txt, """\
route_id,agency_id,route_short_name,route_long_name,route_desc,route_type,\
route_url,route_color,route_text_color
AB,DTA,10,Airport - Bullfrog,Our Airport Route,3,http://example.com,\
00FFFF,000000
""")

    def test_update_geometry_no_trips(self):
        route = Route.objects.create(feed=self.feed, route_id='RTEST', rtype=3)
        self.assertFalse(route.geometry)
        route.update_geometry()
        self.assertFalse(route.geometry)

    def test_update_geometry_1_trip(self):
        route = Route.objects.create(feed=self.feed, route_id='RTEST', rtype=3)
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 3)')
        self.assertFalse(route.geometry)
        route.update_geometry()
        self.assertEqual(route.geometry.coords, (((1.0, 2.0), (1.0, 3.0)),))

    def test_update_geometry_2_trips_different_geometries(self):
        route = Route.objects.create(feed=self.feed, route_id='RTEST', rtype=3)
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 3)')
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 4)')
        self.assertFalse(route.geometry)
        route.update_geometry()
        route_coords = list(route.geometry.coords)
        route_coords.sort()
        self.assertEqual(len(route_coords), 2)
        self.assertEqual(
            route_coords,
            [((1., 2.), (1., 3.)), ((1., 2.), (1., 4.))])

    def test_update_geometry_2_trips_same_geometry(self):
        route = Route.objects.create(feed=self.feed, route_id='RTEST', rtype=3)
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 3)')
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 3)')
        self.assertFalse(route.geometry)
        route.update_geometry()
        route_coords = list(route.geometry.coords)
        route_coords.sort()
        self.assertEqual(len(route_coords), 1)
        self.assertEqual(route.geometry.coords, (((1.0, 2.0), (1.0, 3.0)),))

    def test_update_geometry_no_change(self):
        # For code coverage
        route = Route.objects.create(
            feed=self.feed, route_id='RTEST', rtype=3,
            geometry='MULTILINESTRING((1 2, 1 3))')
        Trip.objects.create(route=route, geometry='LINESTRING(1 2, 1 3)')
        self.assertEqual(route.geometry.coords, (((1.0, 2.0), (1.0, 3.0)),))
        route.update_geometry()
        self.assertEqual(route.geometry.coords, (((1.0, 2.0), (1.0, 3.0)),))



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Service(Base):
    """Dates that a route is active.

    Implements calendar.txt
    """

    feed = models.ForeignKey('Feed')
    service_id = models.CharField(
        max_length=255, db_index=True,
        help_text="Unique identifier for service dates.")
    monday = models.BooleanField(
        default=True,
        help_text="Is the route active on Monday?")
    tuesday = models.BooleanField(
        default=True,
        help_text="Is the route active on Tuesday?")
    wednesday = models.BooleanField(
        default=True,
        help_text="Is the route active on Wednesday?")
    thursday = models.BooleanField(
        default=True,
        help_text="Is the route active on Thursday?")
    friday = models.BooleanField(
        default=True,
        help_text="Is the route active on Friday?")
    saturday = models.BooleanField(
        default=True,
        help_text="Is the route active on Saturday?")
    sunday = models.BooleanField(
        default=True,
        help_text="Is the route active on Sunday?")
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    extra_data = JSONField(default={}, blank=True, null=True)

    def __str__(self):
        return "%d-%s" % (self.feed.id, self.service_id)

    class Meta:
        db_table = 'service'
        app_label = 'multigtfs'

    # For Base import/export
    _column_map = (
        ('service_id', 'service_id'),
        ('monday', 'monday'),
        ('tuesday', 'tuesday'),
        ('wednesday', 'wednesday'),
        ('thursday', 'thursday'),
        ('friday', 'friday'),
        ('saturday', 'saturday'),
        ('sunday', 'sunday'),
        ('start_date', 'start_date'),
        ('end_date', 'end_date')
    )
    _filename = 'calendar.txt'
    _sort_order = ('start_date', 'end_date')
    _unique_fields = ('service_id',)

    @classmethod
    def export_txt(cls, feed):
        '''Export records as calendar.txt'''

        # If no records with start/end dates, skip calendar.txt
        objects = cls.objects.in_feed(feed)

        if not objects.exclude(
                start_date__isnull=True, end_date__isnull=True).exists():
            return None

        return super(Service, cls).export_txt(feed)



#!/usr/bin/env python
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Thanks to django-extensions for the starting code

from __future__ import print_function
import sys

import django
from django.conf import settings

django_version, django_point = django.VERSION[:2]


def base_config():
    '''Create a minimal Django configuration'''
    return {
        'INSTALLED_APPS': ['multigtfs'],
        'TEST_RUNNER': 'django.test.runner.DiscoverRunner',
        'DATABASE_ENGINE': 'django.contrib.gis.db.backends.spatialite',
        'DATABASES': {
            'default': {
                'ENGINE': 'django.contrib.gis.db.backends.spatialite',
            }
        },
        'DEBUG': True,
        'TEMPLATE_DEBUG': True,
        'MIDDLEWARE_CLASSES': '',
    }


def test_config():
    '''Create a Django configuration for running tests'''

    config = base_config()

    # If django-nose is installed, use it
    # You can do things like ./run_tests.py --with-coverage
    try:
        from pkg_resources import WorkingSet, DistributionNotFound
        working_set = WorkingSet()
        working_set.require('django_nose')
    except ImportError:
        print('setuptools not installed.  Weird.')
    except DistributionNotFound:
        print("django-nose not installed.  You'd like it.")
    else:
        config['INSTALLED_APPS'].append('django_nose')
        config['TEST_RUNNER'] = 'django_nose.NoseTestSuiteRunner'

    # Optionally update configuration
    try:
        import t_overrides
    except ImportError:
        pass
    else:
        config = t_overrides.update(config)

    return config


def main(*paths):
    config = test_config()
    settings.configure(**config)
    django.setup()

    from django.core import management
    failures = management.call_command(
        'test',
        *paths
    )
    sys.exit(failures)


if __name__ == '__main__':
    # Extract non-options from command line
    # Options (-sx, --ipdb) will be parsed inside call_command
    paths = []
    for arg in sys.argv[1:]:
        if not arg.startswith('-'):
            paths.append(arg)

    main(*paths)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base
from multigtfs.models.stop import Stop
from multigtfs.models.trip import Trip
from multigtfs.models.fields import SecondsField


@python_2_unicode_compatible
class StopTime(Base):
    """A specific stop on a route on a trip.

    This implements stop_times.txt in the GTFS feed
    """
    trip = models.ForeignKey(Trip)
    stop = models.ForeignKey(Stop)
    arrival_time = SecondsField(
        default=None, null=True, blank=True,
        help_text="Arrival time. Must be set for end stops of trip.")
    departure_time = SecondsField(
        default=None, null=True, blank=True,
        help_text='Departure time. Must be set for end stops of trip.')
    stop_sequence = models.IntegerField()
    stop_headsign = models.CharField(
        max_length=255, blank=True,
        help_text="Sign text that identifies the stop for passengers")
    pickup_type = models.CharField(
        max_length=1, blank=True,
        choices=(('0', 'Regularly scheduled pickup'),
                 ('1', 'No pickup available'),
                 ('2', 'Must phone agency to arrange pickup'),
                 ('3', 'Must coordinate with driver to arrange pickup')),
        help_text="How passengers are picked up")
    drop_off_type = models.CharField(
        max_length=1, blank=True,
        choices=(('0', 'Regularly scheduled drop off'),
                 ('1', 'No drop off available'),
                 ('2', 'Must phone agency to arrange drop off'),
                 ('3', 'Must coordinate with driver to arrange drop off')),
        help_text="How passengers are picked up")
    shape_dist_traveled = models.FloatField(
        "shape distance traveled",
        null=True, blank=True,
        help_text='Distance of stop from start of shape')
    extra_data = JSONField(default={}, blank=True, null=True)

    def __str__(self):
        return "%s-%s-%s" % (self.trip, self.stop.stop_id, self.stop_sequence)

    class Meta:
        db_table = 'stop_time'
        app_label = 'multigtfs'

    _column_map = (
        ('trip_id', 'trip__trip_id'),
        ('arrival_time', 'arrival_time'),
        ('departure_time', 'departure_time'),
        ('stop_id', 'stop__stop_id'),
        ('stop_sequence', 'stop_sequence'),
        ('stop_headsign', 'stop_headsign'),
        ('pickup_type', 'pickup_type'),
        ('drop_off_type', 'drop_off_type'),
        ('shape_dist_traveled', 'shape_dist_traveled')
    )
    _filename = 'stop_times.txt'
    _rel_to_feed = 'trip__route__feed'
    _sort_order = ('trip__trip_id', 'stop_sequence')
    _unique_fields = ('trip_id', 'stop_sequence')



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Fare, FareRule, Route, Zone


class FareRuleTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()
        self.fare = Fare.objects.create(
            feed=self.feed, fare_id='p', price='1.25', currency_type='USD')

    def test_string(self):
        fr = FareRule.objects.create(fare=self.fare)
        self.assertEqual(str(fr), '%d-p' % self.feed.id)
        route = Route.objects.create(feed=self.feed, route_id='R1', rtype=3)
        fr.route = route
        self.assertEqual(str(fr), '%d-p-R1' % self.feed.id)

    def test_import_fare_rules_txt_route(self):
        fare_rules_txt = StringIO("""\
fare_id,route_id,origin_id,destination_id,contains_id
p,AB,,,
""")
        route = Route.objects.create(feed=self.feed, route_id='AB', rtype=3)
        FareRule.import_txt(fare_rules_txt, self.feed)
        fr = FareRule.objects.get()
        self.assertEqual(fr.fare, self.fare)
        self.assertEqual(fr.route, route)
        self.assertEqual(fr.origin, None)
        self.assertEqual(fr.destination, None)
        self.assertEqual(fr.contains, None)

    def test_import_fare_rules_txt_origin(self):
        fare_rules_txt = StringIO("""\
fare_id,route_id,origin_id
p,,2
""")
        zone = Zone.objects.create(feed=self.feed, zone_id='2')
        FareRule.import_txt(fare_rules_txt, self.feed)
        fr = FareRule.objects.get()
        self.assertEqual(fr.fare, self.fare)
        self.assertEqual(fr.route, None)
        self.assertEqual(fr.origin, zone)
        self.assertEqual(fr.destination, None)
        self.assertEqual(fr.contains, None)

    def test_import_fare_rules_txt_full(self):
        fare_rules_txt = StringIO("""\
fare_id,route_id,origin_id,destination_id,contains_id
p,AB,1,2,12
""")
        route = Route.objects.create(feed=self.feed, route_id='AB', rtype=3)
        zone1 = Zone.objects.create(feed=self.feed, zone_id='1')
        zone2 = Zone.objects.create(feed=self.feed, zone_id='2')
        zone12 = Zone.objects.create(feed=self.feed, zone_id='12')
        FareRule.import_txt(fare_rules_txt, self.feed)
        fr = FareRule.objects.get()
        self.assertEqual(fr.fare, self.fare)
        self.assertEqual(fr.route, route)
        self.assertEqual(fr.origin, zone1)
        self.assertEqual(fr.destination, zone2)
        self.assertEqual(fr.contains, zone12)

    def test_import_fare_rules_txt_duplicate(self):
        fare_rules_txt = StringIO("""\
fare_id,route_id
p,AB
p,AB
p,CD
""")
        route = Route.objects.create(feed=self.feed, route_id='AB', rtype=3)
        route2 = Route.objects.create(feed=self.feed, route_id='CD', rtype=3)
        FareRule.import_txt(fare_rules_txt, self.feed)
        self.assertEqual(FareRule.objects.count(), 2)
        fr = FareRule.objects.get(route=route)  # Just one
        self.assertEqual(fr.fare, self.fare)
        fr = FareRule.objects.get(route=route2)
        self.assertEqual(fr.fare, self.fare)

    def test_export_fare_rules_empty(self):
        fare_rules_txt = FareRule.export_txt(self.feed)
        self.assertFalse(fare_rules_txt)

    def test_export_fare_rules_degraded(self):
        # This is possible, but pointless
        FareRule.objects.create(fare=self.fare)
        fare_rules_txt = FareRule.export_txt(self.feed)
        self.assertEqual(fare_rules_txt, '''\
fare_id
p
''')

    def test_export_fare_rules_route_id(self):
        route = Route.objects.create(feed=self.feed, route_id='AB', rtype=3)
        FareRule.objects.create(fare=self.fare, route=route)
        fare_rules_txt = FareRule.export_txt(self.feed)
        self.assertEqual(fare_rules_txt, '''\
fare_id,route_id
p,AB
''')

    def test_export_fare_rules_contains(self):
        zone12 = Zone.objects.create(feed=self.feed, zone_id='12')
        FareRule.objects.create(fare=self.fare, contains=zone12)
        fare_rules_txt = FareRule.export_txt(self.feed)
        self.assertEqual(fare_rules_txt, '''\
fare_id,contains_id
p,12
''')

    def test_export_fare_rules_complete(self):
        route = Route.objects.create(feed=self.feed, route_id='AB', rtype=3)
        zone1 = Zone.objects.create(feed=self.feed, zone_id='1')
        zone2 = Zone.objects.create(feed=self.feed, zone_id='2')
        zone12 = Zone.objects.create(feed=self.feed, zone_id='12')
        FareRule.objects.create(
            fare=self.fare, route=route, origin=zone1, destination=zone2,
            contains=zone12)
        fare_rules_txt = FareRule.export_txt(self.feed)
        self.assertEqual(fare_rules_txt, '''\
fare_id,route_id,origin_id,destination_id,contains_id
p,AB,1,2,12
''')



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from datetime import datetime
import logging

from django.db import connection
from django.conf import settings
from django.core.management.base import BaseCommand

from multigtfs.models import Agency, Feed, Service


class Command(BaseCommand):
    help = 'Import a GTFS Feed'

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument('gtfs_feed',
                            metavar='<gtfsfeed.zip or folder>',
                            type=str)

        # Named (optional) arguments
        parser.add_argument('-n', '--name',
                            type=str,
                            dest='name',
                            help=(
                                'Set the name of the imported feed.  Defaults'
                                ' to name derived from agency name and'
                                ' start date'))

    def handle(self, *args, **options):
        gtfs_feed = options.get('gtfs_feed')
        unset_name = 'Imported at %s' % datetime.now()
        name = options.get('name') or unset_name

        # Setup logging
        verbosity = int(options['verbosity'])
        console = logging.StreamHandler(self.stderr)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        logger_name = 'multigtfs'
        if verbosity == 0:
            level = logging.WARNING
        elif verbosity == 1:
            level = logging.INFO
        elif verbosity == 2:
            level = logging.DEBUG
        else:
            level = logging.DEBUG
            logger_name = ''
            formatter = logging.Formatter(
                '%(name)s - %(levelname)s - %(message)s')
        console.setLevel(level)
        console.setFormatter(formatter)
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.addHandler(console)

        # Disable database query logging
        if settings.DEBUG:
            connection.use_debug_cursor = False

        feed = Feed.objects.create(name=name)
        feed.import_gtfs(gtfs_feed)

        # Set name based on feed
        if feed.name == unset_name:
            try:
                agency = feed.agency_set.order_by('id')[:1].get()
            except Agency.DoesNotExist:
                agency = None
            try:
                service = feed.service_set.order_by('id')[:1].get()
            except Service.DoesNotExist:
                service = None

            if agency:
                name = agency.name
                if service:
                    name += service.start_date.strftime(' starting %Y-%m-%d')
                else:
                    name += ' i' + unset_name[1:]
                feed.name = name
                feed.save()

        self.stdout.write("Successfully imported Feed %s\n" % (feed))



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.contrib.gis import admin

from multigtfs.app_settings import MULTIGTFS_OSMADMIN
from multigtfs.models import (
    Agency, Block, Fare, FareRule, Feed, FeedInfo, Frequency, Route, Service,
    ServiceDate, Shape, ShapePoint, Stop, StopTime, Transfer, Trip, Zone)

geo_admin = admin.OSMGeoAdmin if MULTIGTFS_OSMADMIN else admin.GeoModelAdmin


class AgencyAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


class BlockAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


class FareAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


class FareRuleAdmin(admin.ModelAdmin):
    raw_id_fields = ('fare', 'route', 'origin', 'destination', 'contains')


class FeedInfoAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


class FrequencyAdmin(admin.ModelAdmin):
    raw_id_fields = ('trip', )


class RouteAdmin(geo_admin):
    raw_id_fields = ('feed', 'agency')


class ServiceAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


class ServiceDateAdmin(admin.ModelAdmin):
    raw_id_fields = ('service', )


class ShapeAdmin(geo_admin):
    raw_id_fields = ('feed', )


class ShapePointAdmin(geo_admin):
    raw_id_fields = ('shape', )


class StopAdmin(geo_admin):
    raw_id_fields = ('feed', 'zone', 'parent_station')


class StopTimeAdmin(admin.ModelAdmin):
    raw_id_fields = ('stop', 'trip')


class TransferAdmin(admin.ModelAdmin):
    raw_id_fields = ('from_stop', 'to_stop')


class TripAdmin(geo_admin):
    raw_id_fields = ('route', 'service', 'block', 'shape')


class ZoneAdmin(admin.ModelAdmin):
    raw_id_fields = ('feed', )


admin.site.register(Agency, AgencyAdmin)
admin.site.register(Block, BlockAdmin)
admin.site.register(Fare, FareAdmin)
admin.site.register(FareRule, FareRuleAdmin)
admin.site.register(Feed)
admin.site.register(FeedInfo, FeedInfoAdmin)
admin.site.register(Frequency, FrequencyAdmin)
admin.site.register(Route, RouteAdmin)
admin.site.register(Service, ServiceAdmin)
admin.site.register(ServiceDate, ServiceDateAdmin)
admin.site.register(Shape, ShapeAdmin)
admin.site.register(ShapePoint, ShapePointAdmin)
admin.site.register(Stop, StopAdmin)
admin.site.register(StopTime, StopTimeAdmin)
admin.site.register(Transfer, TransferAdmin)
admin.site.register(Trip, TripAdmin)
admin.site.register(Zone, ZoneAdmin)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Fare


class FareTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()

    def test_string(self):
        fa = Fare.objects.create(
            feed=self.feed, fare_id='p', price='1.25', currency_type='USD',
            payment_method=0, transfers=0)
        self.assertEqual(str(fa), '%d-p(1.25 USD)' % self.feed.id)

    def test_import_fare_attributes_minimal(self):
        fare_attributes_txt = StringIO("""\
fare_id,price,currency_type,payment_method,transfers
p,1.25,USD,0,0
""")
        Fare.import_txt(fare_attributes_txt, self.feed)
        fa = Fare.objects.get()
        self.assertEqual(fa.feed, self.feed)
        self.assertEqual(fa.fare_id, 'p')
        # 1.25 on sqlite, 1.2500 on postgis
        self.assertEqual(str(fa.price)[:4], '1.25')
        self.assertEqual(fa.currency_type, 'USD')
        self.assertEqual(fa.payment_method, 0)
        self.assertEqual(fa.transfers, 0)
        self.assertEqual(fa.transfer_duration, None)

    def test_import_fare_duplicate_fare_id(self):
        fare_attributes_txt = StringIO("""\
fare_id,price,currency_type,payment_method,transfers
p,1.25,USD,0,0
p,1.25,USD,0,0
""")
        Fare.import_txt(fare_attributes_txt, self.feed)
        fa = Fare.objects.get()  # Just one
        self.assertEqual(fa.feed, self.feed)
        self.assertEqual(fa.fare_id, 'p')

    def test_import_fare_attributes_maximal(self):
        fare_attributes_txt = StringIO("""\
fare_id,price,currency_type,payment_method,transfers,transfer_duration
p,1.25,USD,0,0,60
""")
        Fare.import_txt(fare_attributes_txt, self.feed)
        fa = Fare.objects.get()
        self.assertEqual(fa.transfer_duration, 60)

    def test_import_fare_attributes_omitted(self):
        fare_attributes_txt = StringIO("""\
fare_id,price,currency_type,payment_method,transfers,transfer_duration
p,1.25,USD,0,0
""")
        Fare.import_txt(fare_attributes_txt, self.feed)
        fa = Fare.objects.get()
        self.assertEqual(fa.fare_id, 'p')
        self.assertEqual(fa.transfer_duration, None)

    def test_import_fare_attributes_unlimited_transfers(self):
        fare_attributes_txt = StringIO("""\
fare_id,price,currency_type,payment_method,transfers,transfer_duration
p,1.25,USD,0,,3600
""")
        Fare.import_txt(fare_attributes_txt, self.feed)
        fa = Fare.objects.get()
        self.assertEqual(fa.fare_id, 'p')
        self.assertEqual(fa.transfers, None)
        self.assertEqual(fa.transfer_duration, 3600)

    def test_export_fare_attributes_minimal(self):
        Fare.objects.create(
            feed=self.feed, fare_id='p', price='1.25', currency_type='USD',
            payment_method=0, transfers=0)
        fare = Fare.objects.get()
        fare_txt = Fare.export_txt(self.feed)
        self.assertEqual(fare_txt, """\
fare_id,price,currency_type,payment_method,transfers
p,%s,USD,0,0
""" % fare.price)

    def test_export_fare_attributes_maximal(self):
        Fare.objects.create(
            feed=self.feed, fare_id='p', price='1.25', currency_type='USD',
            payment_method=0, transfers=0, transfer_duration=3600)
        fare = Fare.objects.get()
        fare_txt = Fare.export_txt(self.feed)
        self.assertEqual(fare_txt, """\
fare_id,price,currency_type,payment_method,transfers,transfer_duration
p,%s,USD,0,0,3600
""" % fare.price)

    def test_export_fare_attributes_unlimited_transfers(self):
        Fare.objects.create(
            feed=self.feed, fare_id='p', price='1.25', currency_type='USD',
            payment_method=0, transfers=None, transfer_duration=3600)
        fare = Fare.objects.get()
        fare_txt = Fare.export_txt(self.feed)
        self.assertEqual(fare_txt, """\
fare_id,price,currency_type,payment_method,transfers,transfer_duration
p,%s,USD,0,,3600
""" % fare.price)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Transfer(Base):
    """Create additional rules for transfers between ambiguous stops.

    Implements transfer.txt in the GTFS feed.
    """
    from_stop = models.ForeignKey(
        'Stop',
        related_name='transfer_from_stop',
        help_text='Stop where a connection between routes begins.')
    to_stop = models.ForeignKey(
        'Stop',
        related_name='transfer_to_stop',
        help_text='Stop where a connection between routes ends.')
    transfer_type = models.IntegerField(
        default=0, blank=True,
        choices=((0, 'Recommended transfer point'),
                 (1, 'Timed transfer point (vehicle will wait)'),
                 (2, 'min_transfer_time needed to successfully transfer'),
                 (3, 'No transfers possible')),
        help_text="What kind of transfer?")
    min_transfer_time = models.IntegerField(
        null=True, blank=True,
        help_text="How many seconds are required to transfer?")
    extra_data = JSONField(default={}, blank=True, null=True)

    def __str__(self):
        return "%s-%s" % (self.from_stop, self.to_stop.stop_id)

    class Meta:
        db_table = 'transfer'
        app_label = 'multigtfs'

    _column_map = (
        ('from_stop_id', 'from_stop__stop_id'),
        ('to_stop_id', 'to_stop__stop_id'),
        ('transfer_type', 'transfer_type'),
        ('min_transfer_time', 'min_transfer_time')
    )
    _filename = 'transfers.txt'
    _rel_to_feed = 'from_stop__feed'
    _sort_order = ('from_stop__stop_id', 'to_stop__stop_id')
    _unique_fields = ('from_stop_id', 'to_stop_id')



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.contrib.gis.geos import MultiLineString
from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Route, Shape, ShapePoint, Trip


class ShapeTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()

    def test_string(self):
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        self.assertEqual(str(shape), '%d-S1' % self.feed.id)
        shape_pt = ShapePoint.objects.create(
            shape=shape, point="POINT(-117.133162 36.425288)", sequence=1)
        self.assertEqual(str(shape_pt), '%d-S1-1' % self.feed.id)

    def test_legacy_lat_long(self):
        shape = Shape.objects.create(feed=self.feed, shape_id='s1')
        shape_pt1 = ShapePoint(shape=shape, sequence=1)
        shape_pt1.lat = 36.425288
        shape_pt1.lon = -117.133162
        shape_pt1.save()
        shape_pt2 = ShapePoint(shape=shape, sequence=2)
        shape_pt2.lon = -117.14
        shape_pt2.lat = 36.43
        shape_pt2.save()
        self.assertEqual(shape_pt1.point.coords, (-117.133162, 36.425288))
        self.assertEqual(shape_pt1.lat, 36.425288)
        self.assertEqual(shape_pt1.lon, -117.133162)
        self.assertEqual(shape_pt2.point.coords, (-117.14, 36.43))
        self.assertEqual(shape_pt2.lat, 36.43)
        self.assertEqual(shape_pt2.lon, -117.14)

    def test_legacy_create_with_lat_lon(self):
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        shape_pt = ShapePoint.objects.create(
            shape=shape, lat='36.425288', lon='-117.133162', sequence=1)
        self.assertEqual(shape_pt.point.coords, (-117.133162, 36.425288))

    def test_import_shape_minimal(self):
        shape_txt = StringIO("""\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence
S1,36.425288,-117.133162,1
""")
        ShapePoint.import_txt(shape_txt, self.feed)
        shape = Shape.objects.get()
        self.assertEqual(shape.feed, self.feed)
        self.assertEqual(shape.shape_id, 'S1')
        self.assertEqual(shape.geometry, None)
        shape_pt = ShapePoint.objects.get()
        self.assertEqual(shape_pt.shape, shape)
        self.assertEqual(shape_pt.point.coords, (-117.133162, 36.425288))
        self.assertEqual(shape_pt.sequence, 1)
        self.assertEqual(shape_pt.traveled, None)

    def test_import_shape_duplicate(self):
        shape_txt = StringIO("""\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence
S1,36.425288,-117.133162,1
S1,36.42,-117.13,1
S1,36.43,-117.14,2
""")
        ShapePoint.import_txt(shape_txt, self.feed)
        shape = Shape.objects.get()  # Just one
        self.assertEqual(shape.feed, self.feed)
        self.assertEqual(shape.shape_id, 'S1')
        self.assertEqual(2, ShapePoint.objects.count())
        shape_pt, shape_pt2 = ShapePoint.objects.order_by('sequence')
        self.assertEqual(shape_pt.shape, shape)
        self.assertEqual(shape_pt.point.coords, (-117.133162, 36.425288))
        self.assertEqual(shape_pt.sequence, 1)
        self.assertEqual(shape_pt2.shape, shape)
        self.assertEqual(shape_pt2.point.coords, (-117.14, 36.43))
        self.assertEqual(shape_pt2.sequence, 2)

    def test_import_shape_maximal(self):
        shape_txt = StringIO("""\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence,shape_dist_traveled
S1,36.425288,-117.133162,1,0
""")
        ShapePoint.import_txt(shape_txt, self.feed)
        shape = Shape.objects.get()
        self.assertEqual(shape.feed, self.feed)
        self.assertEqual(shape.shape_id, 'S1')
        self.assertEqual(shape.geometry, None)
        shape_pt = ShapePoint.objects.get()
        self.assertEqual(shape_pt.shape, shape)
        self.assertEqual(shape_pt.point.coords, (-117.133162, 36.425288))
        self.assertEqual(shape_pt.sequence, 1)
        self.assertEqual(shape_pt.traveled, 0)

    def test_import_shape_traveled_omitted(self):
        shape_txt = StringIO("""\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence,shape_dist_traveled
S1,36.425288,-117.133162,1,
""")
        ShapePoint.import_txt(shape_txt, self.feed)
        shape = Shape.objects.get()
        self.assertEqual(shape.feed, self.feed)
        self.assertEqual(shape.shape_id, 'S1')
        self.assertEqual(shape.geometry, None)
        shape_pt = ShapePoint.objects.get()
        self.assertEqual(shape_pt.shape, shape)
        self.assertEqual(shape_pt.point.coords, (-117.133162, 36.425288))
        self.assertEqual(shape_pt.sequence, 1)
        self.assertEqual(shape_pt.traveled, None)

    def test_export_shape_empty(self):
        shape_txt = ShapePoint.export_txt(self.feed)
        self.assertFalse(shape_txt)

    def test_export_shape_minimal(self):
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.133162 36.425288)", sequence=1)
        shape_txt = ShapePoint.export_txt(self.feed)
        self.assertEqual(shape_txt, """\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence
S1,36.425288,-117.133162,1
""")

    def test_export_shape_maximal(self):
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.133162 36.425288)", sequence=1,
            traveled=1.1)
        shape_txt = ShapePoint.export_txt(self.feed)
        self.assertEqual(shape_txt, """\
shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence,shape_dist_traveled
S1,36.425288,-117.133162,1,1.1
""")

    def test_update_geometry_on_shapepoint_save(self):
        shape = Shape.objects.create(feed=self.feed)
        route = Route.objects.create(feed=self.feed, rtype=3)
        trip = Trip.objects.create(shape=shape, route=route)
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.133162 36.425288)", sequence=1)
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.13 36.42)", sequence=2)

        shape = Shape.objects.get(id=shape.id)
        trip = Trip.objects.get(id=trip.id)
        route = Route.objects.get(id=route.id)
        self.assertEqual(
            shape.geometry.coords,
            ((-117.133162, 36.425288), (-117.13, 36.42)))
        self.assertEqual(trip.geometry, shape.geometry)
        self.assertEqual(route.geometry,
                         MultiLineString(shape.geometry, srid=4326))

    def test_update_geometry_no_parent(self):
        shape = Shape.objects.create(feed=self.feed)
        route = Route.objects.create(feed=self.feed, rtype=3)
        trip = Trip.objects.create(shape=shape, route=route)
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.133162 36.425288)", sequence=1)
        ShapePoint.objects.create(
            shape=shape, point="POINT(-117.13 36.42)", sequence=2)

        shape.geometry = None
        shape.save()
        trip.geometry = None
        trip.save()
        shape.update_geometry(update_parent=False)

        shape = Shape.objects.get(id=shape.id)
        trip = Trip.objects.get(id=trip.id)
        self.assertEqual(
            shape.geometry.coords,
            ((-117.133162, 36.425288), (-117.13, 36.42)))
        self.assertIsNone(trip.geometry, None)

    def test_shape_geometry_is_ordered(self):
        '''Shape geometry is ordered by ShapePoint sequence

        See issue #24
        '''
        shape = Shape.objects.create(feed=self.feed)
        shape.points.create(point="POINT(-117.2 36.42)", sequence=2)
        shape.points.create(point="POINT(-117.6 36.42)", sequence=6)
        shape.points.create(point="POINT(-117.7 36.42)", sequence=7)
        shape.points.create(point="POINT(-117.3 36.42)", sequence=3)
        shape.points.create(point="POINT(-117.8 36.42)", sequence=8)
        shape.points.create(point="POINT(-117.1 36.42)", sequence=1)
        shape.points.create(point="POINT(-117.4 36.42)", sequence=4)
        shape.points.create(point="POINT(-117.9 36.42)", sequence=9)
        shape.points.create(point="POINT(-117.5 36.42)", sequence=5)
        self.assertEqual(
            shape.geometry.coords,
            ((-117.1, 36.42), (-117.2, 36.42), (-117.3, 36.42),
             (-117.4, 36.42), (-117.5, 36.42), (-117.6, 36.42),
             (-117.7, 36.42), (-117.8, 36.42), (-117.9, 36.42)))



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Agency(Base):
    """One or more transit agencies that provide the data in this feed.

    Maps to agency.txt in the GTFS feed.
    """
    feed = models.ForeignKey('Feed')
    agency_id = models.CharField(
        max_length=255, blank=True, db_index=True,
        help_text="Unique identifier for transit agency")
    name = models.CharField(
        max_length=255,
        help_text="Full name of the transit agency")
    url = models.URLField(
        blank=True, help_text="URL of the transit agency")
    timezone = models.CharField(
        max_length=255,
        help_text="Timezone of the agency")
    lang = models.CharField(
        max_length=2, blank=True,
        help_text="ISO 639-1 code for the primary language")
    phone = models.CharField(
        max_length=255, blank=True,
        help_text="Voice telephone number")
    fare_url = models.URLField(
        blank=True, help_text="URL for purchasing tickets online")
    extra_data = JSONField(default={}, blank=True, null=True)

    def __str__(self):
        return u"%d-%s" % (self.feed.id, self.agency_id)

    class Meta:
        db_table = 'agency'
        app_label = 'multigtfs'
        verbose_name_plural = "agencies"

    # GTFS column names to fields, used by Base for import/export
    _column_map = (
        ('agency_id', 'agency_id'),
        ('agency_name', 'name'),
        ('agency_url', 'url'),
        ('agency_timezone', 'timezone'),
        ('agency_lang', 'lang'),
        ('agency_phone', 'phone'),
        ('agency_fare_url', 'fare_url')
    )
    _filename = 'agency.txt'
    _unique_fields = ('agency_id',)



"""Compatibility utilities.

Handle compatibility between Python versions, Django versions, etc.
"""
from codecs import BOM_UTF8
from distutils.version import LooseVersion
from zipfile import ZipFile, ZIP_DEFLATED

from django import get_version
from django.utils.six import PY3, binary_type, text_type

DJ_VERSION = LooseVersion(get_version())


#
# Get the 'blank' value for a field
#

def _get_blank_value_18(field):
    """Get the value for blank fields in Django 1.8 and earlier."""
    if field.null:
        return None
    else:
        return field.value_to_string(None)


def _get_blank_value_19(field):
    """Get the value for blank fields in Django 1.9 and later."""
    if field.null:
        return None
    else:
        return ''

if DJ_VERSION >= LooseVersion('1.9'):
    get_blank_value = _get_blank_value_19
else:
    get_blank_value = _get_blank_value_18


def bom_prefix_csv(text):
    """
    Prefix CSV text with a Byte-order Marker (BOM).

    The return value needs to be encoded differently so the CSV reader will
    handle the BOM correctly:
    - Python 2 returns a UTF-8 encoded bytestring
    - Python 3 returns unicode text
    """
    if PY3:
        return BOM_UTF8.decode('utf-8') + text
    else:
        return BOM_UTF8 + text.encode('utf-8')


def force_utf8(text):
    """Encode as UTF-8 bytestring if it isn't already."""
    if isinstance(text, binary_type):
        return text
    else:
        return text.encode('utf-8')


def open_writable_zipfile(path):
    """Open a ZipFile for writing, with maximum available compression."""
    try:
        return ZipFile(path, 'w', ZIP_DEFLATED)
    except RuntimeError:  # pragma: nocover
        # zlib module not available
        return ZipFile(path, 'w')


def opener_from_zipfile(zipfile):
    """
    Returns a function that will open a file in a zipfile by name.

    For Python3 compatibility, the raw file will be converted to text.
    """

    def opener(filename):
        inner_file = zipfile.open(filename)
        if PY3:
            from io import TextIOWrapper
            return TextIOWrapper(inner_file)
        else:
            return inner_file

    return opener


def write_text_rows(writer, rows):
    '''Write CSV row data which may include text.'''
    for row in rows:
        try:
            writer.writerow(row)
        except UnicodeEncodeError:
            # Python 2 csv does badly with unicode outside of ASCII
            new_row = []
            for item in row:
                if isinstance(item, text_type):
                    new_row.append(item.encode('utf-8'))
                else:
                    new_row.append(item)
            writer.writerow(new_row)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Fare(Base):
    """A fare class"""

    feed = models.ForeignKey('Feed')
    fare_id = models.CharField(
        max_length=255, db_index=True,
        help_text="Unique identifier for a fare class")
    price = models.DecimalField(
        max_digits=17, decimal_places=4,
        help_text="Fare price, in units specified by currency_type")
    currency_type = models.CharField(
        max_length=3,
        help_text="ISO 4217 alphabetical currency code")
    payment_method = models.IntegerField(
        default=1,
        choices=((0, 'Fare is paid on board.'),
                 (1, 'Fare must be paid before boarding.')),
        help_text="When is the fare paid?")
    transfers = models.IntegerField(
        default=None, null=True, blank=True,
        choices=((0, 'No transfers permitted on this fare.'),
                 (1, 'Passenger may transfer once.'),
                 (2, 'Passenger may transfer twice.'),
                 (None, 'Unlimited transfers are permitted.')),
        help_text="Are transfers permitted?")
    transfer_duration = models.IntegerField(
        null=True, blank=True,
        help_text="Time in seconds until a ticket or transfer expires")
    extra_data = JSONField(default={}, blank=True,  null=True)

    def __str__(self):
        return u"%d-%s(%s %s)" % (
            self.feed.id, self.fare_id, self.price, self.currency_type)

    class Meta:
        db_table = 'fare'
        app_label = 'multigtfs'

    # For Base import/export
    _column_map = (
        ('fare_id', 'fare_id'),
        ('price', 'price'),
        ('currency_type', 'currency_type'),
        ('payment_method', 'payment_method'),
        ('transfers', 'transfers'),
        ('transfer_duration', 'transfer_duration')
    )
    _filename = 'fare_attributes.txt'
    _unique_fields = ('fare_id',)



'''Settings specific to django-multi-gtfs'''
from __future__ import unicode_literals

from django.conf import settings

# If you fulfill the requirements, the OpenStreetMap layer is nicer
# https://docs.djangoproject.com/en/dev/ref/contrib/gis/tutorial/#osmgeoadmin
MULTIGTFS_OSMADMIN = getattr(settings, 'MULTIGTFS_OSMADMIN', True)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from zipfile import ZipFile
import logging
import os
import os.path
import time

from django.contrib.gis.db import models
from django.db.models.signals import post_save
from django.utils.encoding import python_2_unicode_compatible
from django.utils.six import string_types
from jsonfield import JSONField

from multigtfs.compat import open_writable_zipfile, opener_from_zipfile
from .agency import Agency
from .fare import Fare
from .fare_rule import FareRule
from .feed_info import FeedInfo
from .frequency import Frequency
from .route import Route
from .service import Service
from .service_date import ServiceDate
from .shape import ShapePoint, post_save_shapepoint
from .stop import Stop, post_save_stop
from .stop_time import StopTime
from .transfer import Transfer
from .trip import Trip

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class Feed(models.Model):
    """Represents a single GTFS feed.

    This data is not part of the General Transit Feed Specification.  It is
    used to allow storage of several GTFS feeds in the same database.
    """
    name = models.CharField(max_length=255)
    created = models.DateTimeField(auto_now_add=True)
    meta = JSONField(default={}, blank=True, null=True)

    class Meta:
        db_table = 'feed'
        app_label = 'multigtfs'

    def __str__(self):
        if self.name:
            return "%d %s" % (self.id, self.name)
        else:
            return "%d" % self.id

    def import_gtfs(self, gtfs_obj):
        """Import a GTFS file as feed

        Keyword arguments:
        gtfs_obj - A path to a zipped GTFS file, a path to an extracted
            GTFS file, or an open GTFS zip file.

        Returns is a list of objects imported
        """
        total_start = time.time()

        # Determine the type of gtfs_obj
        opener = None
        filelist = None
        if isinstance(gtfs_obj, string_types) and os.path.isdir(gtfs_obj):
            opener = open
            filelist = []
            for dirpath, dirnames, filenames in os.walk(gtfs_obj):
                filelist.extend([os.path.join(dirpath, f) for f in filenames])
        else:
            zfile = ZipFile(gtfs_obj, 'r')
            opener = opener_from_zipfile(zfile)
            filelist = zfile.namelist()

        gtfs_order = (
            Agency, Stop, Route, Service, ServiceDate, ShapePoint, Trip,
            StopTime, Frequency, Fare, FareRule, Transfer, FeedInfo,
        )
        post_save.disconnect(dispatch_uid='post_save_shapepoint')
        post_save.disconnect(dispatch_uid='post_save_stop')
        try:
            for klass in gtfs_order:
                for f in filelist:
                    if f.endswith(klass._filename):
                        start_time = time.time()
                        table = opener(f)
                        count = klass.import_txt(table, self) or 0
                        end_time = time.time()
                        logger.info(
                            'Imported %s (%d %s) in %0.1f seconds',
                            klass._filename, count,
                            klass._meta.verbose_name_plural,
                            end_time - start_time)
                        table.close()

        finally:
            post_save.connect(post_save_shapepoint, sender=ShapePoint)
            post_save.connect(post_save_stop, sender=Stop)

        # Update geometries
        start_time = time.time()
        for shape in self.shape_set.all():
            shape.update_geometry(update_parent=False)
        end_time = time.time()
        logger.info(
            "Updated geometries for %d shapes in %0.1f seconds",
            self.shape_set.count(), end_time - start_time)

        start_time = time.time()
        trips = Trip.objects.in_feed(self)
        for trip in trips:
            trip.update_geometry(update_parent=False)
        end_time = time.time()
        logger.info(
            "Updated geometries for %d trips in %0.1f seconds",
            trips.count(), end_time - start_time)

        start_time = time.time()
        routes = self.route_set.all()
        for route in routes:
            route.update_geometry()
        end_time = time.time()
        logger.info(
            "Updated geometries for %d routes in %0.1f seconds",
            routes.count(), end_time - start_time)

        total_end = time.time()
        logger.info(
            "Import completed in %0.1f seconds.", total_end - total_start)

    def export_gtfs(self, gtfs_file):
        """Export a GTFS file as feed

        Keyword arguments:
        gtfs_file - A path or file-like object for the GTFS feed

        This function will close the file in order to finalize it.
        """
        total_start = time.time()
        z = open_writable_zipfile(gtfs_file)

        gtfs_order = (
            Agency, Service, ServiceDate, Fare, FareRule, FeedInfo, Frequency,
            Route, ShapePoint, StopTime, Stop, Transfer, Trip,
        )

        for klass in gtfs_order:
            start_time = time.time()
            content = klass.export_txt(self)
            if content:
                z.writestr(klass._filename, content)
                end_time = time.time()
                record_count = content.count(type(content)('\n')) - 1
                logger.info(
                    'Exported %s (%d %s) in %0.1f seconds',
                    klass._filename, record_count,
                    klass._meta.verbose_name_plural,
                    end_time - start_time)
        z.close()
        total_end = time.time()
        logger.info(
            'Export completed in %0.1f seconds.', total_end - total_start)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from datetime import date, time

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import (
    Block, Feed, Route, Service, Shape, Stop, StopTime, Trip)


class TripTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()
        self.route = Route.objects.create(
            feed=self.feed, route_id='R1', rtype=3)

    def test_string(self):
        trip = Trip.objects.create(route=self.route, trip_id='T1')
        self.assertEqual(str(trip), '%d-R1-T1' % self.feed.id)

    def test_import_trips_txt_minimal(self):
        trips_txt = StringIO("""\
route_id,service_id,trip_id
R1,S1,T1
""")
        service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        Trip.import_txt(trips_txt, self.feed)
        trip = Trip.objects.get()
        self.assertEqual(trip.route, self.route)
        self.assertEqual(trip.service, service)
        self.assertEqual(trip.trip_id, 'T1')
        self.assertEqual(trip.headsign, '')
        self.assertEqual(trip.short_name, '')
        self.assertEqual(trip.direction, '')
        self.assertEqual(trip.block, None)
        self.assertEqual(trip.shape, None)
        self.assertEqual(trip.wheelchair_accessible, '')
        self.assertEqual(trip.bikes_allowed, '')

    def test_import_trips_txt_duplicate(self):
        trips_txt = StringIO("""\
route_id,service_id,trip_id
R1,S1,T1
R1,S1,T1
""")
        Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        Trip.import_txt(trips_txt, self.feed)
        trip = Trip.objects.get()  # Just one
        self.assertEqual(trip.trip_id, 'T1')

    def test_import_trips_txt_maximal(self):
        trips_txt = StringIO("""\
route_id,service_id,trip_id,trip_headsign,trip_short_name,direction_id,\
block_id,shape_id,wheelchair_accessible,bikes_allowed
R1,S1,T1,Headsign,HS,0,B1,S1,1,2
""")
        service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        block = Block.objects.create(feed=self.feed, block_id='B1')
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        Trip.import_txt(trips_txt, self.feed)
        trip = Trip.objects.get()
        self.assertEqual(trip.route, self.route)
        self.assertEqual(trip.service, service)
        self.assertEqual(trip.trip_id, 'T1')
        self.assertEqual(trip.headsign, 'Headsign')
        self.assertEqual(trip.short_name, 'HS')
        self.assertEqual(trip.direction, '0')
        self.assertEqual(trip.block, block)
        self.assertEqual(trip.shape, shape)
        self.assertEqual(trip.wheelchair_accessible, '1')
        self.assertEqual(trip.bikes_allowed, '2')

    def test_import_trips_txt_multiple_services(self):
        '''
        If a trip is associated with several services, only one is created

        Before 0.4.0, the trip was related to both services
        After 0.4.0, the trip is related to only the first service
        '''
        trips_txt = StringIO("""\
route_id,service_id,trip_id
R1,S1,T1
R1,S2,T1
""")
        service1 = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        service2 = Service.objects.create(
            feed=self.feed, service_id='S2', start_date=date(2012, 1, 1),
            end_date=date(2012, 4, 14))

        Trip.import_txt(trips_txt, self.feed)
        trip = Trip.objects.get()
        self.assertEqual(trip.route, self.route)
        self.assertEqual(trip.service, service1)
        self.assertFalse(service2.trip_set.exists())

    def test_import_trips_txt_quoted_direction_id(self):
        '''
        A direction_id should be stripped of quotation marks

        Issue 64
        '''
        trips_txt = StringIO("""\
route_id,service_id,trip_id,shape_id,trip_headsign,direction_id
R1,"S1","T3","46-860-y11-1.2.I","Aston Quay", "1"
""")
        Trip.import_txt(trips_txt, self.feed)
        trip = Trip.objects.get(trip_id='T3')
        self.assertEqual(trip.direction, '1')

    def test_export_trips_txt_empty(self):
        trips_txt = Trip.export_txt(self.feed)
        self.assertFalse(trips_txt)

    def test_export_trips_txt_minimal(self):
        service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        Trip.objects.create(route=self.route, service=service, trip_id='T1')
        trips_txt = Trip.export_txt(self.feed)
        self.assertEqual(trips_txt, """\
route_id,service_id,trip_id
R1,S1,T1
""")

    def test_export_trips_txt_maximal(self):
        service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        block = Block.objects.create(feed=self.feed, block_id='B1')
        shape = Shape.objects.create(feed=self.feed, shape_id='S1')
        Trip.objects.create(
            route=self.route, service=service, trip_id='T1',
            headsign='Headsign', short_name='HS', direction=0, block=block,
            shape=shape, wheelchair_accessible='2', bikes_allowed='1')
        trips_txt = Trip.export_txt(self.feed)
        self.assertEqual(trips_txt, """\
route_id,service_id,trip_id,trip_headsign,trip_short_name,direction_id,\
block_id,shape_id,wheelchair_accessible,bikes_allowed
R1,S1,T1,Headsign,HS,0,B1,S1,2,1
""")

    def test_update_geometry_no_shape_or_stoptime(self):
        trip = Trip.objects.create(route=self.route, trip_id='T1')
        trip.update_geometry()
        self.assertEqual(trip.geometry, None)

    def test_update_geometry_has_shape(self):
        shape = Shape.objects.create(
            feed=self.feed, shape_id='S1',
            geometry='LINESTRING(-117.133162 36.425288, -117.14 36.43)')
        trip = Trip.objects.create(route=self.route, trip_id='T1', shape=shape)
        trip.update_geometry()
        self.assertEqual(trip.geometry, shape.geometry)

    def test_update_geometry_has_stoptimes(self):
        stop1 = Stop.objects.create(
            feed=self.feed, stop_id='STAGECOACH',
            point="POINT(-117.133162 36.425288)")
        stop2 = Stop.objects.create(
            feed=self.feed, stop_id='TAVERN',
            point="POINT(-117.14 36.43)")
        trip = Trip.objects.create(route=self.route, trip_id='T1')
        StopTime.objects.create(
            trip=trip, stop=stop1, arrival_time=time(6),
            departure_time=time(6), stop_sequence=1)
        StopTime.objects.create(
            trip=trip, stop=stop2, arrival_time=time(7),
            departure_time=time(7), stop_sequence=2)
        trip.update_geometry()
        self.assertEqual(
            trip.geometry.coords,
            ((-117.133162, 36.425288), (-117.14, 36.43)))



# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import unicode_literals

from django.db import models, migrations
import jsonfield.fields
import django.contrib.gis.db.models.fields
import multigtfs.models.fields.seconds


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Agency',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('agency_id', models.CharField(help_text='Unique identifier for transit agency', max_length=255, db_index=True, blank=True)),
                ('name', models.CharField(help_text='Full name of the transit agency', max_length=255)),
                ('url', models.URLField(help_text='URL of the transit agency', blank=True)),
                ('timezone', models.CharField(help_text='Timezone of the agency', max_length=255)),
                ('lang', models.CharField(help_text='ISO 639-1 code for the primary language', max_length=2, blank=True)),
                ('phone', models.CharField(help_text='Voice telephone number', max_length=255, blank=True)),
                ('fare_url', models.URLField(help_text='URL for purchasing tickets online', blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'agency',
                'verbose_name_plural': 'agencies',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Block',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('block_id', models.CharField(help_text='Unique identifier for a block.', max_length=63, db_index=True)),
            ],
            options={
                'db_table': 'block',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Fare',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('fare_id', models.CharField(help_text='Unique identifier for a fare class', max_length=255, db_index=True)),
                ('price', models.DecimalField(help_text='Fare price, in units specified by currency_type', max_digits=17, decimal_places=4)),
                ('currency_type', models.CharField(help_text='ISO 4217 alphabetical currency code', max_length=3)),
                ('payment_method', models.IntegerField(default=1, help_text='When is the fare paid?', choices=[(0, 'Fare is paid on board.'), (1, 'Fare must be paid before boarding.')])),
                ('transfers', models.IntegerField(default=None, help_text='Are transfers permitted?', null=True, blank=True, choices=[(0, 'No transfers permitted on this fare.'), (1, 'Passenger may transfer once.'), (2, 'Passenger may transfer twice.'), (None, 'Unlimited transfers are permitted.')])),
                ('transfer_duration', models.IntegerField(help_text='Time in seconds until a ticket or transfer expires', null=True, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'fare',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FareRule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'fare_rules',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Feed',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('meta', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'feed',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='FeedInfo',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('publisher_name', models.CharField(help_text='Full name of organization that publishes the feed.', max_length=255)),
                ('publisher_url', models.URLField(help_text="URL of the feed publisher's organization.")),
                ('lang', models.CharField(help_text='IETF BCP 47 language code for text in field.', max_length=20, verbose_name='language')),
                ('start_date', models.DateField(help_text='Date that feed starts providing reliable data.', null=True, blank=True)),
                ('end_date', models.DateField(help_text='Date that feed stops providing reliable data.', null=True, blank=True)),
                ('version', models.CharField(help_text='Version of feed.', max_length=255, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
            ],
            options={
                'db_table': 'feed_info',
                'verbose_name_plural': 'feed info',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Frequency',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('start_time', multigtfs.models.fields.seconds.SecondsField(help_text='Time that the service begins at the specified frequency')),
                ('end_time', multigtfs.models.fields.seconds.SecondsField(help_text='Time that the service ends at the specified frequency')),
                ('headway_secs', models.IntegerField(help_text='Time in seconds before returning to same stop')),
                ('exact_times', models.CharField(blank=True, help_text='Should frequency-based trips be exactly scheduled?', max_length=1, choices=[(0, 'Trips are not exactly scheduled'), (1, 'Trips are exactly scheduled from start time')])),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
            ],
            options={
                'db_table': 'frequency',
                'verbose_name_plural': 'frequencies',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Route',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('route_id', models.CharField(help_text='Unique identifier for route.', max_length=255, db_index=True)),
                ('short_name', models.CharField(help_text='Short name of the route', max_length=63)),
                ('long_name', models.CharField(help_text='Long name of the route', max_length=255)),
                ('desc', models.TextField(help_text='Long description of a route', verbose_name='description', blank=True)),
                ('rtype', models.IntegerField(help_text='Type of transportation used on route', verbose_name='route type', choices=[(0, 'Tram, Streetcar, or Light rail'), (1, 'Subway or Metro'), (2, 'Rail'), (3, 'Bus'), (4, 'Ferry'), (5, 'Cable car'), (6, 'Gondola or Suspended cable car'), (7, 'Funicular')])),
                ('url', models.URLField(help_text='Web page about for the route', blank=True)),
                ('color', models.CharField(help_text='Color of route in hex', max_length=6, blank=True)),
                ('text_color', models.CharField(help_text='Color of route text in hex', max_length=6, blank=True)),
                ('geometry', django.contrib.gis.db.models.fields.MultiLineStringField(help_text='Geometry cache of Trips', srid=4326, null=True, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('agency', models.ForeignKey(blank=True, to='multigtfs.Agency', help_text='Agency for this route.', null=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
            ],
            options={
                'db_table': 'route',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Service',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('service_id', models.CharField(help_text='Unique identifier for service dates.', max_length=255, db_index=True)),
                ('monday', models.BooleanField(default=True, help_text='Is the route active on Monday?')),
                ('tuesday', models.BooleanField(default=True, help_text='Is the route active on Tuesday?')),
                ('wednesday', models.BooleanField(default=True, help_text='Is the route active on Wednesday?')),
                ('thursday', models.BooleanField(default=True, help_text='Is the route active on Thursday?')),
                ('friday', models.BooleanField(default=True, help_text='Is the route active on Friday?')),
                ('saturday', models.BooleanField(default=True, help_text='Is the route active on Saturday?')),
                ('sunday', models.BooleanField(default=True, help_text='Is the route active on Sunday?')),
                ('start_date', models.DateField(null=True, blank=True)),
                ('end_date', models.DateField(null=True, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
            ],
            options={
                'db_table': 'service',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ServiceDate',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateField(help_text='Date that the service differs from the norm.')),
                ('exception_type', models.IntegerField(default=1, help_text='Is service added or removed on this date?', choices=[(1, 'Added'), (2, 'Removed')])),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('service', models.ForeignKey(to='multigtfs.Service')),
            ],
            options={
                'db_table': 'service_date',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Shape',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('shape_id', models.CharField(help_text='Unique identifier for a shape.', max_length=255, db_index=True)),
                ('geometry', django.contrib.gis.db.models.fields.LineStringField(help_text='Geometry cache of ShapePoints', srid=4326, null=True, blank=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
            ],
            options={
                'db_table': 'shape',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='ShapePoint',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('point', django.contrib.gis.db.models.fields.PointField(help_text='WGS 84 latitude/longitude of shape point', srid=4326)),
                ('sequence', models.IntegerField()),
                ('traveled', models.FloatField(help_text='Distance of point from start of shape', null=True, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('shape', models.ForeignKey(related_name='points', to='multigtfs.Shape')),
            ],
            options={
                'db_table': 'shape_point',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Stop',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('stop_id', models.CharField(help_text='Unique identifier for a stop or station.', max_length=255, db_index=True)),
                ('code', models.CharField(help_text='Uniquer identifier (short text or number) for passengers.', max_length=255, blank=True)),
                ('name', models.CharField(help_text='Name of stop in local vernacular.', max_length=255)),
                ('desc', models.CharField(help_text='Description of a stop.', max_length=255, verbose_name='description', blank=True)),
                ('point', django.contrib.gis.db.models.fields.PointField(help_text='WGS 84 latitude/longitude of stop or station', srid=4326)),
                ('url', models.URLField(help_text='URL for the stop', blank=True)),
                ('location_type', models.CharField(blank=True, help_text='Is this a stop or station?', max_length=1, choices=[('0', 'Stop'), ('1', 'Station')])),
                ('timezone', models.CharField(help_text='Timezone of the stop', max_length=255, blank=True)),
                ('wheelchair_boarding', models.CharField(blank=True, help_text='Is wheelchair boarding possible?', max_length=1, choices=[('0', 'No information'), ('1', 'Some wheelchair boarding'), ('2', 'No wheelchair boarding')])),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
                ('parent_station', models.ForeignKey(blank=True, to='multigtfs.Stop', help_text='The station associated with the stop', null=True)),
            ],
            options={
                'db_table': 'stop',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='StopTime',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('arrival_time', multigtfs.models.fields.seconds.SecondsField(default=None, help_text='Arrival time. Must be set for end stops of trip.', null=True, blank=True)),
                ('departure_time', multigtfs.models.fields.seconds.SecondsField(default=None, help_text='Departure time. Must be set for end stops of trip.', null=True, blank=True)),
                ('stop_sequence', models.IntegerField()),
                ('stop_headsign', models.CharField(help_text='Sign text that identifies the stop for passengers', max_length=255, blank=True)),
                ('pickup_type', models.CharField(blank=True, help_text='How passengers are picked up', max_length=1, choices=[('0', 'Regularly scheduled pickup'), ('1', 'No pickup available'), ('2', 'Must phone agency to arrange pickup'), ('3', 'Must coordinate with driver to arrange pickup')])),
                ('drop_off_type', models.CharField(blank=True, help_text='How passengers are picked up', max_length=1, choices=[('0', 'Regularly scheduled drop off'), ('1', 'No drop off available'), ('2', 'Must phone agency to arrange drop off'), ('3', 'Must coordinate with driver to arrange drop off')])),
                ('shape_dist_traveled', models.FloatField(help_text='Distance of stop from start of shape', null=True, verbose_name='shape distance traveled', blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('stop', models.ForeignKey(to='multigtfs.Stop')),
            ],
            options={
                'db_table': 'stop_time',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Transfer',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('transfer_type', models.IntegerField(default=0, help_text='What kind of transfer?', blank=True, choices=[(0, 'Recommended transfer point'), (1, 'Timed transfer point (vehicle will wait)'), (2, 'min_transfer_time needed to successfully transfer'), (3, 'No transfers possible')])),
                ('min_transfer_time', models.IntegerField(help_text='How many seconds are required to transfer?', null=True, blank=True)),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('from_stop', models.ForeignKey(related_name='transfer_from_stop', to='multigtfs.Stop', help_text='Stop where a connection between routes begins.')),
                ('to_stop', models.ForeignKey(related_name='transfer_to_stop', to='multigtfs.Stop', help_text='Stop where a connection between routes ends.')),
            ],
            options={
                'db_table': 'transfer',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Trip',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('trip_id', models.CharField(help_text='Unique identifier for a trip.', max_length=255, db_index=True)),
                ('headsign', models.CharField(help_text='Destination identification for passengers.', max_length=255, blank=True)),
                ('short_name', models.CharField(help_text='Short name used in schedules and signboards.', max_length=63, blank=True)),
                ('direction', models.CharField(blank=True, help_text='Direction for bi-directional routes.', max_length=1, choices=[('0', '0'), ('1', '1')])),
                ('geometry', django.contrib.gis.db.models.fields.LineStringField(help_text='Geometry cache of Shape or Stops', srid=4326, null=True, blank=True)),
                ('wheelchair_accessible', models.CharField(blank=True, help_text='Are there accommodations for riders with wheelchair?', max_length=1, choices=[('0', 'No information'), ('1', 'Some wheelchair accommodation'), ('2', 'No wheelchair accommodation')])),
                ('bikes_allowed', models.CharField(blank=True, help_text='Are bicycles allowed?', max_length=1, choices=[('0', 'No information'), ('1', 'Some bicycle accommodation'), ('2', 'No bicycles allowed')])),
                ('extra_data', jsonfield.fields.JSONField(default={}, null=True, blank=True)),
                ('block', models.ForeignKey(blank=True, to='multigtfs.Block', help_text='Block of sequential trips that this trip belongs to.', null=True)),
                ('route', models.ForeignKey(to='multigtfs.Route')),
                ('service', models.ForeignKey(blank=True, to='multigtfs.Service', null=True)),
                ('shape', models.ForeignKey(blank=True, to='multigtfs.Shape', help_text='Shape used for this trip', null=True)),
            ],
            options={
                'db_table': 'trip',
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Zone',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('zone_id', models.CharField(help_text='Unique identifier for a zone.', max_length=63, db_index=True)),
                ('feed', models.ForeignKey(to='multigtfs.Feed')),
            ],
            options={
                'db_table': 'zone',
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='stoptime',
            name='trip',
            field=models.ForeignKey(to='multigtfs.Trip'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='stop',
            name='zone',
            field=models.ForeignKey(blank=True, to='multigtfs.Zone', help_text='Fare zone for a stop ID.', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='frequency',
            name='trip',
            field=models.ForeignKey(to='multigtfs.Trip'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='farerule',
            name='contains',
            field=models.ForeignKey(related_name='fare_contains', blank=True, to='multigtfs.Zone', help_text='Fare class is valid for travel withing this zone.', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='farerule',
            name='destination',
            field=models.ForeignKey(related_name='fare_destinations', blank=True, to='multigtfs.Zone', help_text='Fare class is valid for travel ending in this zone.', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='farerule',
            name='fare',
            field=models.ForeignKey(to='multigtfs.Fare'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='farerule',
            name='origin',
            field=models.ForeignKey(related_name='fare_origins', blank=True, to='multigtfs.Zone', help_text='Fare class is valid for travel originating in this zone.', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='farerule',
            name='route',
            field=models.ForeignKey(blank=True, to='multigtfs.Route', help_text='Fare class is valid for this route.', null=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='fare',
            name='feed',
            field=models.ForeignKey(to='multigtfs.Feed'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='block',
            name='feed',
            field=models.ForeignKey(to='multigtfs.Feed'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='agency',
            name='feed',
            field=models.ForeignKey(to='multigtfs.Feed'),
            preserve_default=True,
        ),
    ]



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Agency, Feed
from multigtfs.compat import bom_prefix_csv


class AgencyTest(TestCase):

    def setUp(self):
        self.feed = Feed.objects.create()

    def test_string(self):
        agency = Agency.objects.create(feed=self.feed, agency_id='TEST')
        self.assertEqual(str(agency), '%d-TEST' % self.feed.id)

    def test_import_minimal(self):
        agency_txt = StringIO("""\
agency_name,agency_url,agency_timezone
Demo Transit Authority,http://google.com,America/Los_Angeles
""")
        Agency.import_txt(agency_txt, self.feed)
        agency = Agency.objects.get()
        self.assertEqual(agency.agency_id, '')
        self.assertEqual(agency.name, 'Demo Transit Authority')
        self.assertEqual(agency.url, 'http://google.com')
        self.assertEqual(agency.timezone, 'America/Los_Angeles')
        self.assertEqual(agency.lang, '')
        self.assertEqual(agency.phone, '')
        self.assertEqual(agency.fare_url, '')

    def test_import_two_no_id(self):
        agency_txt = StringIO("""\
agency_name,agency_url,agency_timezone
Demo Transit Authority,http://google.com,America/Los_Angeles
Example Transit Authority,http://example.com,America/Los_Angeles
""")
        Agency.import_txt(agency_txt, self.feed)
        agency = Agency.objects.get()  # Just one
        self.assertEqual(agency.agency_id, '')
        self.assertEqual(agency.name, 'Demo Transit Authority')

    def test_import_two_same_id(self):
        agency_txt = StringIO("""\
agency_id,agency_name,agency_url,agency_timezone
DTA,Demo Transit Authority,http://google.com,America/Los_Angeles
DTA,Example Transit Authority,http://example.com,America/Los_Angeles
""")
        Agency.import_txt(agency_txt, self.feed)
        agency = Agency.objects.get()  # Just one
        self.assertEqual(agency.agency_id, 'DTA')
        self.assertEqual(agency.name, 'Demo Transit Authority')

    def test_import_agency_maximal(self):
        agency_txt = StringIO("""\
agency_id,agency_name,agency_url,agency_timezone,agency_lang,agency_phone,\
agency_fare_url
DTA,"Demo Transit Authority",http://google.com,America/Los_Angeles,en,\
555-555-TEST,http://google.com
""")
        Agency.import_txt(agency_txt, self.feed)
        agency = Agency.objects.get()
        self.assertEqual(agency.agency_id, 'DTA')
        self.assertEqual(agency.name, 'Demo Transit Authority')
        self.assertEqual(agency.url, 'http://google.com')
        self.assertEqual(agency.timezone, 'America/Los_Angeles')
        self.assertEqual(agency.lang, 'en')
        self.assertEqual(agency.phone, '555-555-TEST')
        self.assertEqual(agency.fare_url, 'http://google.com')

    def test_import_bom(self):
        agency_txt = StringIO(bom_prefix_csv("""\
agency_name,agency_url,agency_timezone
Demo Transit Authority,http://google.com,America/Los_Angeles
"""))
        Agency.import_txt(agency_txt, self.feed)
        agency = Agency.objects.get()
        self.assertEqual(agency.agency_id, '')
        self.assertEqual(agency.name, 'Demo Transit Authority')
        self.assertEqual(agency.url, 'http://google.com')
        self.assertEqual(agency.timezone, 'America/Los_Angeles')
        self.assertEqual(agency.lang, '')
        self.assertEqual(agency.phone, '')
        self.assertEqual(agency.fare_url, '')

    def test_export_agency_none(self):
        agency_txt = Agency.export_txt(self.feed)
        self.assertFalse(agency_txt)

    def test_export_agency_minimal(self):
        Agency.objects.create(
            feed=self.feed, name='Demo Transit Authority',
            url='http://google.com', timezone='America/Los_Angeles')
        agency_txt = Agency.export_txt(self.feed)
        self.assertEqual(agency_txt, """\
agency_name,agency_url,agency_timezone
Demo Transit Authority,http://google.com,America/Los_Angeles
""")

    def test_export_agency_maximal(self):
        Agency.objects.create(
            feed=self.feed, agency_id='DTA', name='Demo Transit Authority',
            url='http://google.com', timezone='America/Los_Angeles',
            lang='en', phone='555-555-TEST', fare_url='http://google.com')
        agency_txt = Agency.export_txt(self.feed)
        self.assertEqual(agency_txt, """\
agency_id,agency_name,agency_url,agency_timezone,agency_lang,agency_phone,\
agency_fare_url
DTA,Demo Transit Authority,http://google.com,America/Los_Angeles,en,\
555-555-TEST,http://google.com
""")



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
import logging
import time

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.db import connection

from multigtfs.models import Feed, Route, Shape, Trip


class Command(BaseCommand):
    help = 'Updates the cached geometry of GTFS feeds'

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument('feed_ids',
                            nargs='*',
                            metavar='Feed ID',
                            type=int)

        # Named (optional) arguments
        parser.add_argument('-a', '--all',
                            action='store_true',
                            dest='all',
                            default=False,
                            help='Update all feeds')
        parser.add_argument('-q', '--quiet',
                            action='store_false',
                            dest='verbose',
                            default=True,
                            help="Don't print status messages to stdout")

    def handle(self, *args, **options):
        total_start = time.time()

        # Validate the arguments
        all_feeds = options.get('all')
        feed_ids = options.get('feed_ids')
        if len(feed_ids) == 0 and not all_feeds:
            raise CommandError('You must pass in a feed ID or --all.')
        if len(feed_ids) > 0 and all_feeds:
            raise CommandError("You can't specify a feed and --all.")

        # Setup logging
        verbosity = int(options['verbosity'])
        console = logging.StreamHandler(self.stderr)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        logger_name = 'multigtfs'
        if verbosity == 0:
            level = logging.WARNING
        elif verbosity == 1:
            level = logging.INFO
        elif verbosity == 2:
            level = logging.DEBUG
        else:
            level = logging.DEBUG
            logger_name = ''
            formatter = logging.Formatter(
                '%(name)s - %(levelname)s - %(message)s')
        console.setLevel(level)
        console.setFormatter(formatter)
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.addHandler(console)

        # Disable database query logging
        if settings.DEBUG:
            connection.use_debug_cursor = False

        # Get the feeds
        if all_feeds:
            feeds = Feed.objects.order_by('id')
        else:
            feeds = []
            for feed_id in feed_ids:
                try:
                    feeds.append(Feed.objects.get(id=feed_id))
                except Feed.DoesNotExist:
                    raise CommandError('Feed %s not found' % feed_id)

        # Refresh the geometries
        for feed in feeds:
            logger.info(
                "Updating geometries in Feed %s (ID %s)...",
                feed.name, feed.id)

            start_time = time.time()
            shapes = Shape.objects.in_feed(feed)
            for shape in shapes:
                shape.update_geometry(update_parent=False)
            end_time = time.time()
            logger.debug(
                "Imported %s shape%s in %0.1f seconds",
                shapes.count(), '' if shapes.count() == 1 else 's',
                end_time - start_time)

            start_time = time.time()
            trips = Trip.objects.in_feed(feed)
            for trip in trips:
                trip.update_geometry(update_parent=False)
            end_time = time.time()
            logger.debug(
                "Imported %s trip%s in %0.1f seconds",
                trips.count(), '' if trips.count() == 1 else 's',
                end_time - start_time)

            start_time = time.time()
            routes = Route.objects.in_feed(feed)
            for route in routes:
                route.update_geometry()
            end_time = time.time()
            logger.debug(
                "Imported %s route%s in %0.1f seconds",
                routes.count(), '' if routes.count() == 1 else 's',
                end_time - start_time)

            total_end = time.time()
            logger.info(
                "Feed %d: Updated geometries in %d shape%s, %d trip%s, and"
                " %d route%s %0.1f seconds.",
                feed.id,
                shapes.count(), '' if shapes.count() == 1 else 's',
                trips.count(), '' if trips.count() == 1 else 's',
                routes.count(), '' if routes.count() == 1 else 's',
                total_end - total_start)



# -*- coding: utf-8 -*-
# Copyright 2012-2014 John Whitlock
# Copyright 2014 Juha Yrjölä
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
from codecs import BOM_UTF8
from collections import defaultdict
from csv import reader, writer
from datetime import datetime, date
from logging import getLogger
import re

from django.contrib.gis.db import models
from django.contrib.gis.db.models.query import GeoQuerySet
from django.db.models.fields.related import ManyToManyField
from django.utils.six import StringIO, text_type, PY3

from multigtfs.compat import get_blank_value, write_text_rows

logger = getLogger(__name__)
re_point = re.compile(r'(?P<name>point)\[(?P<index>\d)\]')
batch_size = 1000
large_queryset_size = 100000
CSV_BOM = BOM_UTF8.decode('utf-8') if PY3 else BOM_UTF8


class BaseQuerySet(GeoQuerySet):
    def populated_column_map(self):
        '''Return the _column_map without unused optional fields'''
        column_map = []
        cls = self.model
        for csv_name, field_pattern in cls._column_map:
            # Separate the local field name from foreign columns
            if '__' in field_pattern:
                field_name = field_pattern.split('__', 1)[0]
            else:
                field_name = field_pattern

            # Handle point fields
            point_match = re_point.match(field_name)
            if point_match:
                field = None
            else:
                field = cls._meta.get_field(field_name)

            # Only add optional columns if they are used in the records
            if field and field.blank and not field.has_default():
                kwargs = {field_name: get_blank_value(field)}
                if self.exclude(**kwargs).exists():
                    column_map.append((csv_name, field_pattern))
            else:
                column_map.append((csv_name, field_pattern))
        return column_map


class BaseManager(models.GeoManager):
    def get_queryset(self):
        '''Return the custom queryset.'''
        return BaseQuerySet(self.model)

    def in_feed(self, feed):
        '''Return the objects in the target feed'''
        kwargs = {self.model._rel_to_feed: feed}
        return self.filter(**kwargs)


class Base(models.Model):
    """Base class for models that are defined in the GTFS spec

    Implementers need to define a class variable:

    _column_map - A mapping of GTFS columns to model fields
    It should be set to a sequence of tuples:
    - GTFS column name
    - Model field name

    If the column is optional, then set blank=True on the field, and set
    null=True appropriately.

    Implementers can define this class variable:

    _rel_to_feed - The relation of this model to the field, in Django filter
    format.  The default is 'feed', and will be used to get the objects
    on a feed like this:
    Model.objects.filter(_rel_to_feed=feed)

    """

    class Meta:
        abstract = True
        app_label = 'multigtfs'

    objects = BaseManager()

    # The relation of the model to the feed it belongs to.
    _rel_to_feed = 'feed'

    @classmethod
    def import_txt(cls, txt_file, feed, filter_func=None):
        '''Import from the GTFS text file'''

        # Setup the conversion from GTFS to Django Format
        # Conversion functions
        def no_convert(value): return value

        def date_convert(value): return datetime.strptime(value, '%Y%m%d')

        def bool_convert(value): return (value == '1')

        def char_convert(value): return (value or '')

        def null_convert(value): return (value or None)

        def point_convert(value): return (value or 0.0)

        cache = {}

        def default_convert(field):
            def get_value_or_default(value):
                if value == '' or value is None:
                    return field.get_default()
                else:
                    return value
            return get_value_or_default

        def instance_convert(field, feed, rel_name):
            def get_instance(value):
                if value.strip():
                    key1 = "{}:{}".format(field.rel.to.__name__, rel_name)
                    key2 = text_type(value)

                    # Load existing objects
                    if key1 not in cache:
                        pairs = field.rel.to.objects.filter(
                            **{field.rel.to._rel_to_feed: feed}).values_list(
                            rel_name, 'id')
                        cache[key1] = dict((text_type(x), i) for x, i in pairs)

                    # Create new?
                    if key2 not in cache[key1]:
                        kwargs = {
                            field.rel.to._rel_to_feed: feed,
                            rel_name: value}
                        cache[key1][key2] = field.rel.to.objects.create(
                            **kwargs).id
                    return cache[key1][key2]
                else:
                    return None
            return get_instance

        # Check unique fields
        column_names = [c for c, _ in cls._column_map]
        for unique_field in cls._unique_fields:
            assert unique_field in column_names, \
                '{} not in {}'.format(unique_field, column_names)

        # Map of field_name to converters from GTFS to Django format
        val_map = dict()
        name_map = dict()
        point_map = dict()
        for csv_name, field_pattern in cls._column_map:
            # Separate the local field name from foreign columns
            if '__' in field_pattern:
                field_base, rel_name = field_pattern.split('__', 1)
                field_name = field_base + '_id'
            else:
                field_name = field_base = field_pattern
            # Use the field name in the name mapping
            name_map[csv_name] = field_name

            # Is it a point field?
            point_match = re_point.match(field_name)
            if point_match:
                field = None
            else:
                field = cls._meta.get_field(field_base)

            # Pick a conversion function for the field
            if point_match:
                converter = point_convert
            elif isinstance(field, models.DateField):
                converter = date_convert
            elif isinstance(field, models.BooleanField):
                converter = bool_convert
            elif isinstance(field, models.CharField):
                converter = char_convert
            elif field.rel:
                converter = instance_convert(field, feed, rel_name)
                assert not isinstance(field, models.ManyToManyField)
            elif field.null:
                converter = null_convert
            elif field.has_default():
                converter = default_convert(field)
            else:
                converter = no_convert

            if point_match:
                index = int(point_match.group('index'))
                point_map[csv_name] = (index, converter)
            else:
                val_map[csv_name] = converter

        # Read and convert the source txt
        csv_reader = reader(txt_file, skipinitialspace=True)
        unique_line = dict()
        count = 0
        first = True
        extra_counts = defaultdict(int)
        new_objects = []
        for row in csv_reader:
            if first:
                # Read the columns
                columns = row
                if columns[0].startswith(CSV_BOM):
                    columns[0] = columns[0][len(CSV_BOM):]
                first = False
                continue

            if filter_func and not filter_func(zip(columns, row)):
                continue

            if not row:
                continue

            # Read a data row
            fields = dict()
            point_coords = [None, None]
            ukey_values = {}
            if cls._rel_to_feed == 'feed':
                fields['feed'] = feed
            for column_name, value in zip(columns, row):
                if column_name not in name_map:
                    val = null_convert(value)
                    if val is not None:
                        fields.setdefault('extra_data', {})[column_name] = val
                        extra_counts[column_name] += 1
                elif column_name in val_map:
                    fields[name_map[column_name]] = val_map[column_name](value)
                else:
                    assert column_name in point_map
                    pos, converter = point_map[column_name]
                    point_coords[pos] = converter(value)

                # Is it part of the unique key?
                if column_name in cls._unique_fields:
                    ukey_values[column_name] = value

            # Join the lat/long into a point
            if point_map:
                assert point_coords[0] and point_coords[1]
                fields['point'] = "POINT(%s)" % (' '.join(point_coords))

            # Is the item unique?
            ukey = tuple(ukey_values.get(u) for u in cls._unique_fields)
            if ukey in unique_line:
                logger.warning(
                    '%s line %d is a duplicate of line %d, not imported.',
                    cls._filename, csv_reader.line_num, unique_line[ukey])
                continue
            else:
                unique_line[ukey] = csv_reader.line_num

            # Create after accumulating a batch
            new_objects.append(cls(**fields))
            if len(new_objects) % batch_size == 0:  # pragma: no cover
                cls.objects.bulk_create(new_objects)
                count += len(new_objects)
                logger.info(
                    "Imported %d %s",
                    count, cls._meta.verbose_name_plural)
                new_objects = []

        # Create remaining objects
        if new_objects:
            cls.objects.bulk_create(new_objects)

        # Take note of extra fields
        if extra_counts:
            extra_columns = feed.meta.setdefault(
                'extra_columns', {}).setdefault(cls.__name__, [])
            for column in columns:
                if column in extra_counts and column not in extra_columns:
                    extra_columns.append(column)
            feed.save()
        return len(unique_line)

    @classmethod
    def export_txt(cls, feed):
        '''Export records as a GTFS comma-separated file'''
        objects = cls.objects.in_feed(feed)

        # If no records, return None
        if not objects.exists():
            return

        # Get the columns used in the dataset
        column_map = objects.populated_column_map()
        columns, fields = zip(*column_map)
        extra_columns = feed.meta.get(
            'extra_columns', {}).get(cls.__name__, [])

        # Get sort order
        if hasattr(cls, '_sort_order'):
            sort_fields = cls._sort_order
        else:
            sort_fields = []
            for field in fields:
                base_field = field.split('__', 1)[0]
                point_match = re_point.match(base_field)
                if point_match:
                    continue
                field_type = cls._meta.get_field(base_field)
                assert not isinstance(field_type, ManyToManyField)
                sort_fields.append(field)

        # Create CSV writer
        out = StringIO()
        csv_writer = writer(out, lineterminator='\n')

        # Write header row
        header_row = [text_type(c) for c in columns]
        header_row.extend(extra_columns)
        write_text_rows(csv_writer, [header_row])

        # Report the work to be done
        total = objects.count()
        logger.info(
            '%d %s to export...',
            total, cls._meta.verbose_name_plural)

        # Populate related items cache
        model_to_field_name = {}
        cache = {}
        for field_name in fields:
            if '__' in field_name:
                local_field_name, subfield_name = field_name.split('__', 1)
                field = cls._meta.get_field(local_field_name)
                field_type = field.rel.to
                model_name = field_type.__name__
                if model_name in model_to_field_name:
                    # Already loaded this model under a different field name
                    cache[field_name] = cache[model_to_field_name[model_name]]
                else:
                    # Load all feed data for this model
                    pairs = field_type.objects.in_feed(
                        feed).values_list('id', subfield_name)
                    cache[field_name] = dict(
                        (i, text_type(x)) for i, x in pairs)
                    cache[field_name][None] = u''
                    model_to_field_name[model_name] = field_name

        # For large querysets, break up by the first field
        if total < large_queryset_size:
            querysets = [objects.order_by(*sort_fields)]
        else:  # pragma: no cover
            field1_raw = sort_fields[0]
            assert '__' in field1_raw
            assert field1_raw in cache
            field1 = field1_raw.split('__', 1)[0]
            field1_id = field1 + '_id'

            # Sort field1 ids by field1 values
            val_to_id = dict((v, k) for k, v in cache[field1_raw].items())
            assert len(val_to_id) == len(cache[field1_raw])
            sorted_vals = sorted(val_to_id.keys())

            querysets = []
            for val in sorted_vals:
                fid = val_to_id[val]
                if fid:
                    qs = objects.filter(
                        **{field1_id: fid}).order_by(*sort_fields[1:])
                    querysets.append(qs)

        # Assemble the rows, writing when we hit batch size
        count = 0
        rows = []
        for queryset in querysets:
            for item in queryset.order_by(*sort_fields):
                row = []
                for csv_name, field_name in column_map:
                    obj = item
                    point_match = re_point.match(field_name)
                    if '__' in field_name:
                        # Return relations from cache
                        local_field_name = field_name.split('__', 1)[0]
                        field_id = getattr(obj, local_field_name + '_id')
                        row.append(cache[field_name][field_id])
                    elif point_match:
                        # Get the lat or long from the point
                        name, index = point_match.groups()
                        field = getattr(obj, name)
                        row.append(field.coords[int(index)])
                    else:
                        # Handle other field types
                        field = getattr(obj, field_name) if obj else ''
                        if isinstance(field, date):
                            formatted = field.strftime(u'%Y%m%d')
                            row.append(text_type(formatted))
                        elif isinstance(field, bool):
                            row.append(1 if field else 0)
                        elif field is None:
                            row.append(u'')
                        else:
                            row.append(text_type(field))
                for col in extra_columns:
                    row.append(obj.extra_data.get(col, u''))
                rows.append(row)
                if len(rows) % batch_size == 0:  # pragma: no cover
                    write_text_rows(csv_writer, rows)
                    count += len(rows)
                    logger.info(
                        "Exported %d %s",
                        count, cls._meta.verbose_name_plural)
                    rows = []

        # Write rows smaller than batch size
        write_text_rows(csv_writer, rows)
        return out.getvalue()



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals

from django.test import TestCase

from multigtfs.models import Feed, Zone


class ZoneTest(TestCase):
    def test_string(self):
        feed = Feed.objects.create()
        zone = Zone.objects.create(feed=feed, zone_id='ZONE')
        self.assertEqual(str(zone), '%d-ZONE' % feed.id)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from datetime import time

from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Route, Stop, StopTime, Trip


class StopTimeTest(TestCase):

    def setUp(self):
        self.feed = Feed.objects.create()
        self.route = Route.objects.create(
            feed=self.feed, route_id='R1', rtype=3)
        self.trip = Trip.objects.create(route=self.route, trip_id='STBA')
        self.stop = Stop.objects.create(
            feed=self.feed, stop_id='STAGECOACH',
            point="POINT(-117.133162 36.425288)")

    def test_string(self):
        stoptime = StopTime.objects.create(
            trip=self.trip, stop=self.stop, arrival_time=time(6),
            departure_time=time(6), stop_sequence=1)
        self.assertEqual(
            str(stoptime), '%d-R1-STBA-STAGECOACH-1' % self.feed.id)

    def test_import_stop_times_txt_minimal(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence
STBA,6:00:00,6:00:00,STAGECOACH,1
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime.departure_time), '06:00:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.stop_headsign, '')
        self.assertEqual(stoptime.pickup_type, '')
        self.assertEqual(stoptime.drop_off_type, '')
        self.assertEqual(stoptime.shape_dist_traveled, None)

    def test_import_stop_times_txt_duplicate(self):
        Stop.objects.create(
            feed=self.feed, stop_id='XXX',
            point="POINT(-117.133 36.425)")
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence
STBA,6:00:00,6:00:00,STAGECOACH,1
STBA,7:00:00,7:00:00,XXX,1
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()  # Just one
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')

    def test_import_stop_times_txt_bad_column_empty_OK(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,drop_off_time
STBA,6:00:00,6:00:00,STAGECOACH,1,
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime.departure_time), '06:00:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.drop_off_type, '')
        self.assertEqual(stoptime.extra_data, {})

    def test_import_stop_times_txt_bad_column_populated_OK(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,drop_off_time
STBA,6:00:00,6:00:00,STAGECOACH,1,1
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime.departure_time), '06:00:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.drop_off_type, '')
        self.assertEqual(stoptime.extra_data, {'drop_off_time': '1'})

    def test_import_stop_times_txt_maximal(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,stop_headsign,\
pickup_type,drop_off_type,shape_dist_traveled
STBA,6:00:00,6:00:00,STAGECOACH,1,"SC",2,1,5.25
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime.departure_time), '06:00:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.stop_headsign, 'SC')
        self.assertEqual(stoptime.pickup_type, '2')
        self.assertEqual(stoptime.drop_off_type, '1')
        self.assertEqual(stoptime.shape_dist_traveled, 5.25)

    def test_import_stop_times_txt_empty_optional(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,stop_headsign,\
pickup_type,drop_off_type,shape_dist_traveled
STBA,6:00:00,6:00:00,STAGECOACH,1,,,,
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime.departure_time), '06:00:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.stop_headsign, '')
        self.assertEqual(stoptime.pickup_type, '')
        self.assertEqual(stoptime.drop_off_type, '')
        self.assertEqual(stoptime.shape_dist_traveled, None)

    def test_import_stop_times_txt_middle_times_optional(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence
STBA,6:00:00,6:00:00,STAGECOACH,1
STBA,,,STAGECOACH2,2
STBA,12:00:00,12:00:00,STAGECOACH3,3
""")
        stop2 = Stop.objects.create(
            feed=self.feed, stop_id='STAGECOACH2',
            point="POINT(-117.133162 36.425288)")
        stop3 = Stop.objects.create(
            feed=self.feed, stop_id='STAGECOACH3',
            point="POINT(-117.133162 36.425288)")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime1 = StopTime.objects.get(stop=self.stop)
        self.assertEqual(str(stoptime1.arrival_time), '06:00:00')
        self.assertEqual(str(stoptime1.departure_time), '06:00:00')
        stoptime2 = StopTime.objects.get(stop=stop2)
        self.assertEqual(stoptime2.arrival_time, None)
        self.assertEqual(stoptime2.departure_time, None)
        stoptime3 = StopTime.objects.get(stop=stop3)
        self.assertEqual(str(stoptime3.arrival_time), '12:00:00')
        self.assertEqual(str(stoptime3.departure_time), '12:00:00')

    def test_import_stop_times_txt_tomorrow(self):
        stop_times_txt = StringIO("""\
trip_id,arrival_time,departure_time,stop_id,stop_sequence
STBA,23:59:00,24:01:00,STAGECOACH,1
""")
        StopTime.import_txt(stop_times_txt, self.feed)
        stoptime = StopTime.objects.get()
        self.assertEqual(stoptime.trip, self.trip)
        self.assertEqual(str(stoptime.arrival_time), '23:59:00')
        self.assertEqual(str(stoptime.departure_time), '24:01:00')
        self.assertEqual(stoptime.stop, self.stop)
        self.assertEqual(stoptime.stop_sequence, 1)
        self.assertEqual(stoptime.stop_headsign, '')
        self.assertEqual(stoptime.pickup_type, '')
        self.assertEqual(stoptime.drop_off_type, '')
        self.assertEqual(stoptime.shape_dist_traveled, None)

    def test_export_stop_times_none(self):
        stop_times_txt = StopTime.export_txt(self.feed)
        self.assertFalse(stop_times_txt)

    def test_export_stop_times_minimal(self):
        StopTime.objects.create(
            trip=self.trip, arrival_time='6:00:00', departure_time='6:00:00',
            stop=self.stop, stop_sequence=1)
        stop_times_txt = StopTime.export_txt(self.feed)
        self.assertEqual(stop_times_txt, """\
trip_id,arrival_time,departure_time,stop_id,stop_sequence
STBA,06:00:00,06:00:00,STAGECOACH,1
""")

    def test_export_stop_times_maximal(self):
        StopTime.objects.create(
            trip=self.trip, arrival_time='6:00:00', departure_time='6:00:00',
            stop=self.stop, stop_sequence=1, stop_headsign='SC',
            pickup_type=2, drop_off_type=1, shape_dist_traveled=5.25)
        stop_times_txt = StopTime.export_txt(self.feed)
        self.assertEqual(stop_times_txt, """\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,stop_headsign,\
pickup_type,drop_off_type,shape_dist_traveled
STBA,06:00:00,06:00:00,STAGECOACH,1,SC,2,1,5.25
""")

    def test_export_natural_sort(self):
        StopTime.objects.create(
            trip=self.trip, arrival_time='6:00:00', departure_time='6:00:00',
            stop=self.stop, stop_sequence=1, stop_headsign='SC',
            pickup_type=2, drop_off_type=1, shape_dist_traveled=5.25)
        stop2 = Stop.objects.create(
            feed=self.feed, stop_id='SALOON', point="POINT(-117.1 36.5)")
        StopTime.objects.create(trip=self.trip, stop=stop2, stop_sequence=2)
        stop3 = Stop.objects.create(
            feed=self.feed, stop_id='GENERAL_STORE',
            point="POINT(-117.2 36.5)")
        StopTime.objects.create(trip=self.trip, stop=stop3, stop_sequence=3)
        stop4 = Stop.objects.create(
            feed=self.feed, stop_id='MORGUE', point="POINT(-117.2 36.6)")
        StopTime.objects.create(
            trip=self.trip, arrival_time='7:00:00', departure_time='7:00:00',
            stop=stop4, stop_sequence=4, stop_headsign='MORT')
        stop_times_txt = StopTime.export_txt(self.feed)
        self.assertEqual(stop_times_txt, """\
trip_id,arrival_time,departure_time,stop_id,stop_sequence,stop_headsign,\
pickup_type,drop_off_type,shape_dist_traveled
STBA,06:00:00,06:00:00,STAGECOACH,1,SC,2,1,5.25
STBA,,,SALOON,2,,,,
STBA,,,GENERAL_STORE,3,,,,
STBA,07:00:00,07:00:00,MORGUE,4,MORT,,,
""")



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals
import logging

from django.db import connection
from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.template.defaultfilters import slugify

from multigtfs.models.feed import Feed


class Command(BaseCommand):
    help = 'Exports a GTFS Feed to a zipped feed file'

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument('feed_id',
                            metavar='GTFS Feed ID',
                            type=int)

        # Named (optional) arguments
        parser.add_argument('-n', '--name',
                            type=str,
                            dest='name',
                            help='Set the name of the exported feed')

    def handle(self, *args, **options):
        # Setup logging
        verbosity = int(options['verbosity'])
        console = logging.StreamHandler(self.stderr)
        formatter = logging.Formatter('%(levelname)s - %(message)s')
        logger_name = 'multigtfs'
        if verbosity == 0:
            level = logging.WARNING
        elif verbosity == 1:
            level = logging.INFO
        elif verbosity == 2:
            level = logging.DEBUG
        else:
            level = logging.DEBUG
            logger_name = ''
            formatter = logging.Formatter(
                '%(name)s - %(levelname)s - %(message)s')
        console.setLevel(level)
        console.setFormatter(formatter)
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)
        logger.addHandler(console)

        # Disable database query logging
        if settings.DEBUG:
            connection.use_debug_cursor = False

        feed_id = options.get('feed_id')

        try:
            feed = Feed.objects.get(id=feed_id)
        except Feed.DoesNotExist:
            raise CommandError('Feed %s not found' % feed_id)
        out_name = options.get('name') or slugify(feed.name)
        if not out_name.endswith('.zip'):
            out_name += '.zip'
        self.stdout.write(
            "Exporting Feed %s to %s...\n" % (feed_id, out_name))
        feed.export_gtfs(out_name)
        self.stdout.write(
            "Successfully exported Feed %s to %s\n" % (feed_id, out_name))



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.contrib.gis.geos import LineString
from django.utils.encoding import python_2_unicode_compatible
from jsonfield import JSONField

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Trip(Base):
    """A trip along a route

    This implements trips.txt in the GTFS feed
    """

    route = models.ForeignKey('Route')
    service = models.ForeignKey('Service', null=True, blank=True)
    trip_id = models.CharField(
        max_length=255, db_index=True,
        help_text="Unique identifier for a trip.")
    headsign = models.CharField(
        max_length=255, blank=True,
        help_text="Destination identification for passengers.")
    short_name = models.CharField(
        max_length=63, blank=True,
        help_text="Short name used in schedules and signboards.")
    direction = models.CharField(
        max_length=1, blank=True,
        choices=(('0', '0'), ('1', '1')),
        help_text="Direction for bi-directional routes.")
    block = models.ForeignKey(
        'Block', null=True, blank=True,
        help_text="Block of sequential trips that this trip belongs to.")
    shape = models.ForeignKey(
        'Shape', null=True, blank=True,
        help_text="Shape used for this trip")
    geometry = models.LineStringField(
        null=True, blank=True,
        help_text='Geometry cache of Shape or Stops')
    wheelchair_accessible = models.CharField(
        max_length=1, blank=True,
        choices=(
            ('0', 'No information'),
            ('1', 'Some wheelchair accommodation'),
            ('2', 'No wheelchair accommodation')),
        help_text='Are there accommodations for riders with wheelchair?')
    bikes_allowed = models.CharField(
        max_length=1, blank=True,
        choices=(
            ('0', 'No information'),
            ('1', 'Some bicycle accommodation'),
            ('2', 'No bicycles allowed')),
        help_text='Are bicycles allowed?')
    extra_data = JSONField(default={}, blank=True, null=True)

    def update_geometry(self, update_parent=True):
        """Update the geometry from the Shape or Stops"""
        original = self.geometry
        if self.shape:
            self.geometry = self.shape.geometry
        else:
            stoptimes = self.stoptime_set.order_by('stop_sequence')
            if stoptimes.count() > 1:
                self.geometry = LineString(
                    [st.stop.point.coords for st in stoptimes])
        if self.geometry != original:
            self.save()
            if update_parent:
                self.route.update_geometry()

    def __str__(self):
        return "%s-%s" % (self.route, self.trip_id)

    class Meta:
        db_table = 'trip'
        app_label = 'multigtfs'

    _column_map = (
        ('route_id', 'route__route_id'),
        ('service_id', 'service__service_id'),
        ('trip_id', 'trip_id'),
        ('trip_headsign', 'headsign'),
        ('trip_short_name', 'short_name'),
        ('direction_id', 'direction'),
        ('block_id', 'block__block_id'),
        ('shape_id', 'shape__shape_id'),
        ('wheelchair_accessible', 'wheelchair_accessible'),
        ('bikes_allowed', 'bikes_allowed'),
    )
    _filename = 'trips.txt'
    _rel_to_feed = 'route__feed'
    _unique_fields = ('trip_id',)



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Define model and field to represent time of day in GTFS feeds"

from __future__ import unicode_literals

from django.db.models import Field
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Seconds(object):
    '''A GTFS seconds value, formatted as HH:MM:SS in the GTFS feed'''

    def __init__(self, seconds=0):
        self.seconds = int(seconds)
        if self.seconds < 0:
            raise ValueError('seconds must be positive')

    @classmethod
    def from_hms(cls, hours=0, minutes=0, seconds=0):
        return Seconds((hours * 60 * 60) + (minutes * 60) + seconds)

    def __str__(self):
        minutes, seconds = divmod(self.seconds, 60)
        hours, minutes = divmod(minutes, 60)
        return "%02d:%02d:%02d" % (hours, minutes, seconds)

    def _compare(self, other, method):
        try:
            return method(self.seconds, other.seconds)
        except (AttributeError, TypeError):
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s, o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s, o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s, o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s, o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s, o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s, o: s != o)


class SecondsField(Field):
    '''A Model Field for storing Seconds'''

    description = 'Seconds since start of the day'

    def from_db_value(self, value, expression, connection, context):
        '''Handle data loaded from database.'''
        if value is None:
            return value
        return self.parse_seconds(value)

    def to_python(self, value):
        '''Handle data from serialization and form clean() methods.'''
        if isinstance(value, Seconds):
            return value
        if value in self.empty_values:
            return None
        return self.parse_seconds(value)

    @staticmethod
    def parse_seconds(value):
        '''
        Parse string into Seconds instances.

        Handled formats:
        HH:MM:SS
        HH:MM
        SS
        '''
        svalue = str(value)
        colons = svalue.count(':')
        if colons == 2:
            hours, minutes, seconds = [int(v) for v in svalue.split(':')]
        elif colons == 1:
            hours, minutes = [int(v) for v in svalue.split(':')]
            seconds = 0
        elif colons == 0:
            hours = 0
            minutes = 0
            seconds = int(svalue)
        else:
            raise ValueError('Must be in seconds or HH:MM:SS format')
        return Seconds.from_hms(hours, minutes, seconds)

    def get_prep_value(self, value):
        '''Prepare value for database storage.'''
        if isinstance(value, Seconds):
            return value.seconds
        elif value:
            return self.parse_seconds(value).seconds
        else:
            return None

    def get_internal_type(self):
        '''Seconds are stored in the database like nullable Integers.'''
        return 'IntegerField'

    def value_to_string(self, obj):
        '''Convert to HH:MM:SS format.'''
        value = self.value_from_object(obj)
        return value.__str__()



from django.views.generic import ListView
from multigtfs.models import (
    Block, Fare, FareRule, Feed, Frequency, Route, Service, ServiceDate, Shape,
    ShapePoint, Stop, StopTime, Trip)


class ByFeedListView(ListView):
    by_col = 'feed_id'
    by_kwarg = 'feed_id'
    by_class = Feed
    by_classname = 'feed'

    def get_context_data(self, **kwargs):
        context = super(ByFeedListView, self).get_context_data(
            **kwargs)
        context[self.by_classname] = self.by_class.objects.get(
            id=self.kwargs[self.by_kwarg])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        q_filter = {self.by_col: self.kwargs[self.by_kwarg]}
        qset = super(ByFeedListView, self).get_queryset(**kwargs)
        return qset.filter(**q_filter)


class FareRuleByFareListView(ListView):
    model = FareRule

    def get_context_data(self, **kwargs):
        context = super(FareRuleByFareListView, self).get_context_data(
            **kwargs)
        context['fare'] = Fare.objects.get(id=self.kwargs['fare_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return FareRule.objects.filter(fare_id=self.kwargs['fare_id'])


class FareRuleByRouteListView(ListView):
    model = FareRule

    def get_context_data(self, **kwargs):
        context = super(FareRuleByRouteListView, self).get_context_data(
            **kwargs)
        context['route'] = Route.objects.get(id=self.kwargs['route_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return FareRule.objects.filter(route_id=self.kwargs['route_id'])


class FrequencyByTripListView(ListView):
    model = Frequency

    def get_context_data(self, **kwargs):
        context = super(FrequencyByTripListView, self).get_context_data(
            **kwargs)
        context['trip'] = Trip.objects.get(id=self.kwargs['trip_id'])
        return context

    def get_queryset(self, **kwargs):
        return Frequency.objects.filter(trip=self.kwargs['trip_id'])


class ServiceDateByServiceListView(ListView):
    model = ServiceDate

    def get_context_data(self, **kwargs):
        context = super(ServiceDateByServiceListView, self).get_context_data(
            **kwargs)
        context['service'] = Service.objects.get(id=self.kwargs['service_id'])
        return context

    def get_queryset(self, **kwargs):
        return ServiceDate.objects.filter(service=self.kwargs['service_id'])


class ShapePointByShapeListView(ListView):
    model = ShapePoint

    def get_context_data(self, **kwargs):
        context = super(ShapePointByShapeListView, self).get_context_data(
            **kwargs)
        context['shape'] = Shape.objects.get(id=self.kwargs['shape_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return ShapePoint.objects.filter(shape=self.kwargs['shape_id'])


class StopTimeByStopListView(ListView):
    model = StopTime

    def get_context_data(self, **kwargs):
        context = super(StopTimeByStopListView, self).get_context_data(
            **kwargs)
        context['stop'] = Stop.objects.get(id=self.kwargs['stop_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return StopTime.objects.filter(stop_id=self.kwargs['stop_id'])


class StopTimeByTripListView(ListView):
    model = StopTime

    def get_context_data(self, **kwargs):
        context = super(StopTimeByTripListView, self).get_context_data(
            **kwargs)
        context['trip'] = Trip.objects.get(id=self.kwargs['trip_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return StopTime.objects.filter(trip=self.kwargs['trip_id'])


class TripByBlockListView(ListView):
    model = Trip

    def get_context_data(self, **kwargs):
        context = super(TripByBlockListView, self).get_context_data(**kwargs)
        context['the_block'] = Block.objects.get(id=self.kwargs['block_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return Trip.objects.filter(block_id=self.kwargs['block_id'])


class TripByRouteListView(ListView):
    model = Trip

    def get_context_data(self, **kwargs):
        context = super(TripByRouteListView, self).get_context_data(**kwargs)
        context['route'] = Route.objects.get(id=self.kwargs['route_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return Trip.objects.filter(route_id=self.kwargs['route_id'])


class TripByServiceListView(ListView):
    model = Trip

    def get_context_data(self, **kwargs):
        context = super(TripByServiceListView, self).get_context_data(**kwargs)
        context['service'] = Service.objects.get(id=self.kwargs['service_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return Trip.objects.filter(service=self.kwargs['service_id'])


class TripByShapeListView(ListView):
    model = ShapePoint

    def get_context_data(self, **kwargs):
        context = super(TripByShapeListView, self).get_context_data(**kwargs)
        context['shape'] = Shape.objects.get(id=self.kwargs['shape_id'])
        context['feed_id'] = self.kwargs['feed_id']
        return context

    def get_queryset(self, **kwargs):
        return Trip.objects.filter(shape=self.kwargs['shape_id'])



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import unicode_literals

from django.utils.encoding import python_2_unicode_compatible

from multigtfs.models.base import models, Base


@python_2_unicode_compatible
class Zone(Base):
    """Represents a fare zone.

    This data is not represented as a file in the GTFS.  It appears as an
    identifier in the fare_rules and the stop tables.
    """
    feed = models.ForeignKey('Feed')
    zone_id = models.CharField(
        max_length=63, db_index=True,
        help_text="Unique identifier for a zone.")

    def __str__(self):
        return "%d-%s" % (self.feed_id, self.zone_id)

    class Meta:
        db_table = 'zone'
        app_label = 'multigtfs'



#
# Copyright 2012-2014 John Whitlock
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from datetime import date
from json import loads

from django.core.serializers import serialize
from django.test import TestCase
from django.utils.six import StringIO

from multigtfs.models import Feed, Frequency, Route, Service, Trip
from multigtfs.models.fields import Seconds


class FrequencyTest(TestCase):
    def setUp(self):
        self.feed = Feed.objects.create()
        self.route = Route.objects.create(
            feed=self.feed, route_id='R1', rtype=3)
        self.service = Service.objects.create(
            feed=self.feed, service_id='S1', start_date=date(2011, 4, 14),
            end_date=date(2011, 12, 31))
        self.trip = Trip.objects.create(
            route=self.route, service=self.service, trip_id='STBA')

    def test_string(self):
        frequency = Frequency.objects.create(
            trip=self.trip, start_time='6:00', end_time='22:00',
            headway_secs=1800)
        self.assertEqual(str(frequency), '%d-R1-STBA' % self.feed.id)

    def test_import_frequencies_txt_minimal(self):
        frequencies_txt = StringIO("""\
trip_id,start_time,end_time,headway_secs
STBA,6:00:00,22:00:00,1800
""")
        Frequency.import_txt(frequencies_txt, self.feed)
        frequency = Frequency.objects.get()
        self.assertEqual(frequency.trip, self.trip)
        self.assertEqual(frequency.start_time, Seconds.from_hms(hours=6))
        self.assertEqual(frequency.end_time, Seconds.from_hms(hours=22))
        self.assertEqual(frequency.headway_secs, 1800)
        self.assertEqual(frequency.exact_times, '')

    def test_import_frequencies_txt_duplicate(self):
        frequencies_txt = StringIO("""\
trip_id,start_time,end_time,headway_secs
STBA,6:00:00,8:00:00,1800
STBA,6:00:00,10:00:00,1200
STBA,10:00:00,12:00:00,1500
""")
        Frequency.import_txt(frequencies_txt, self.feed)
        self.assertEqual(2, Frequency.objects.count())
        freq1, freq2 = Frequency.objects.order_by('start_time')
        self.assertEqual(freq1.trip, self.trip)
        self.assertEqual(freq1.start_time, Seconds.from_hms(hours=6))
        self.assertEqual(freq1.end_time, Seconds.from_hms(hours=8))
        self.assertEqual(freq1.headway_secs, 1800)
        self.assertEqual(freq2.trip, self.trip)
        self.assertEqual(freq2.start_time, Seconds.from_hms(hours=10))
        self.assertEqual(freq2.end_time, Seconds.from_hms(hours=12))
        self.assertEqual(freq2.headway_secs, 1500)

    def test_import_frequencies_txt_maximal(self):
        frequencies_txt = StringIO("""\
trip_id,start_time,end_time,headway_secs,exact_times
STBA,6:00:00,23:30:35,1800,1
""")
        Frequency.import_txt(frequencies_txt, self.feed)
        frequency = Frequency.objects.get()
        self.assertEqual(frequency.trip, self.trip)
        self.assertEqual(frequency.start_time, Seconds.from_hms(hours=6))
        self.assertEqual(frequency.end_time, Seconds.from_hms(23, 30, 35))
        self.assertEqual(frequency.headway_secs, 1800)
        self.assertEqual(frequency.exact_times, '1')

    def test_import_frequencies_txt_omitted_with_rollover(self):
        frequencies_txt = StringIO("""\
trip_id,start_time,end_time,headway_secs,exact_times
STBA,00:50:00,24:10:00,1800,
""")
        Frequency.import_txt(frequencies_txt, self.feed)
        frequency = Frequency.objects.get()
        self.assertEqual(str(frequency.start_time), '00:50:00')
        self.assertEqual(frequency.end_time, Seconds.from_hms(24, 10))
        self.assertEqual(frequency.headway_secs, 1800)
        self.assertEqual(frequency.exact_times, '')

    def test_export_frequencies_txt_none(self):
        frequencies_txt = Frequency.export_txt(self.feed)
        self.assertEqual(frequencies_txt, None)

    def test_export_frequencies_txt_minimal(self):
        Frequency.objects.create(
            trip=self.trip, start_time=Seconds.from_hms(hours=6),
            end_time=Seconds.from_hms(hours=22), headway_secs=1800)
        frequencies_txt = Frequency.export_txt(self.feed)
        self.assertEqual(frequencies_txt, """\
trip_id,start_time,end_time,headway_secs
STBA,06:00:00,22:00:00,1800
""")

    def test_export_frequencies_txt_maximal(self):
        Frequency.objects.create(
            trip=self.trip, start_time='05:00', end_time='25:00',
            headway_secs=1800)
        frequencies_txt = Frequency.export_txt(self.feed)
        self.assertEqual(frequencies_txt, """\
trip_id,start_time,end_time,headway_secs
STBA,05:00:00,25:00:00,1800
""")

    def test_serialize(self):
        '''Test serialization of Frequency, which has a SecondsField'''
        f = Frequency.objects.create(
            trip=self.trip, start_time='05:00', end_time='25:00',
            headway_secs=1800)
        actual = loads(serialize('json', Frequency.objects.all()))
        expected = [{
            u"pk": f.id,
            u"model": u"multigtfs.frequency",
            u"fields": {
                u"exact_times": u"",
                u"extra_data": u"{}",
                u"start_time": u"05:00:00",
                u"headway_secs": 1800,
                u"trip": self.trip.id,
                u"end_time": u"25:00:00"}}]
        self.maxDiff = None
        self.assertEqual(expected, actual)
